/*
 * bg.c
 *
 * Last modified: <2009/11/23 23:05:09 +0900> By SUMIDA
 */
#include <stdio.h>
#include <stdarg.h>
#include <SDL.h>
#include "system.h"
#include "surface.h"
#include "bg.h"
#include "texture.h"

#define OPT_USE_ATTRIBUTE 0x01
#define OPT_DEFAULT (OPT_USE_ATTRIBUTE)

BG bg[BG_MAX];
FONTMAP fontmap;

void bg_init(void) {

	int i;

	printf("bg_init ... ");

	for (i = 0; i < BG_MAX; i++) {
		BG *p = &bg[i];
		//p->priority = i * 256 / BG_MAX;
		p->ofs_x   = 8;
		p->ofs_y   = 8;

		memset(bg[i].map, 0, sizeof(bg[i].map));
	}

	printf("done.\n");
}

void bg_rendaring(SDL_Surface *screen, int bg_no) {

	BG *p = &bg[bg_no];
	int col, row;
	int ofs_x, ofs_y;
	int cmax, rmax;
	int co, ro;

	p->ofs_x %= (BG_MAP_WIDTH  + BG_MAP_RESERVE) * BG_MAP_SIZE;
	p->ofs_y %= (BG_MAP_HEIGHT + BG_MAP_RESERVE) * BG_MAP_SIZE;
	col = p->ofs_x / BG_MAP_SIZE;
	row = p->ofs_y / BG_MAP_SIZE;
	ofs_x = col * BG_MAP_SIZE - p->ofs_x;
	ofs_y = row * BG_MAP_SIZE - p->ofs_y;
	cmax = (ofs_x == 0)? BG_MAP_WIDTH : BG_MAP_WIDTH  + 1;
	rmax = (ofs_y == 0)? BG_MAP_HEIGHT: BG_MAP_HEIGHT + 1;

	if (bg_no == 0) {

	for (ro = 0; ro < rmax; ro++) {
		int r = (BG_MAP_HEIGHT + BG_MAP_RESERVE + row + ro) % (BG_MAP_HEIGHT + BG_MAP_RESERVE);
		int y = ofs_y + (ro << 4);

		for (co = 0; co < cmax; co++) {
			int c = (BG_MAP_WIDTH + BG_MAP_RESERVE + col + co) % (BG_MAP_WIDTH + BG_MAP_RESERVE);
			MAP_ATTR *attr = &p->map[r][c];
			if (attr->map_id) {
				int x = ofs_x + (co << 4);
				surface_draw_frame16(screen, attr->surface_id, attr->map_id, x, y);
			}
		}
	}

		return;
	}

	for (ro = 0; ro < rmax; ro++) {
		int r = (BG_MAP_HEIGHT + BG_MAP_RESERVE + row + ro) % (BG_MAP_HEIGHT + BG_MAP_RESERVE);
		int y = ofs_y + ro * BG_MAP_SIZE;

		for (co = 0; co < cmax; co++) {
			int c = (BG_MAP_WIDTH + BG_MAP_RESERVE + col + co) % (BG_MAP_WIDTH + BG_MAP_RESERVE);
			MAP_ATTR *attr = &p->map[r][c];
#if 0
			if (c < 0 || c >= BG_MAP_WIDTH + BG_MAP_RESERVE
				|| r < 0 || r >= BG_MAP_HEIGHT + BG_MAP_RESERVE) {
				fprintf(stderr, "p->ofs_x=%d, p->ofs_y=%d, c=%d, r=%d, row=%d, ro=%d, col=%d, co=%d\n", p->ofs_x, p->ofs_y, c, r, row, ro, col, co);
			}
#endif
			if (attr->map_id) {
				int x = ofs_x + co * BG_MAP_SIZE;
#if (BG_MAP_SIZE == 8)
				surface_draw_frame8(screen, attr->surface_id, attr->map_id, x, y);
#elif (BG_MAP_SIZE == 16)
				surface_draw_frame16(screen, attr->surface_id, attr->map_id, x, y);
#elif (BG_MAP_SIZE == 32)
				surface_draw_frame32(screen, attr->surface_id, attr->map_id, x, y);
#else
				surface_draw_frame(screen, attr->surface_id, attr->map_id, x, y);
#endif
			}
		}
	}

}


void font_init(void) {

	int i;

	printf("font_init ... ");

	fontmap.default_tex_id = TEX_FONT;
	fontmap.tex_id = TEX_FONT;
	fontmap.ofs_x   = 8;
	fontmap.ofs_y   = 8;
	fontmap.font_x  = 0;
	fontmap.font_y  = 0;
	fontmap.frame_x = 1;
	fontmap.frame_y = 1;
	fontmap.frame_w = FONT_MAP_WIDTH;
	fontmap.frame_h = FONT_MAP_HEIGHT;
	memset(fontmap.map, 0, sizeof(fontmap.map));

	printf("done.\n");
}

void font_clean(void) {
	memset(fontmap.map, 0, sizeof(fontmap.map));
	fontmap.tex_id = fontmap.default_tex_id;
	fontmap.options = OPT_DEFAULT;
}

void font_texture(int tex_id) {
	fontmap.tex_id = tex_id;
}

void font_default_texture(void) {
	fontmap.tex_id = fontmap.default_tex_id;
}

void font_rendaring(SDL_Surface *screen) {

	FONTMAP *p = &fontmap;
	int col, row;
	int ofs_x, ofs_y;
	int cmax, rmax;
	int co, ro;

	p->ofs_x %= (FONT_MAP_WIDTH  + FONT_MAP_RESERVE) * FONT_MAP_SIZE;
	p->ofs_y %= (FONT_MAP_HEIGHT + FONT_MAP_RESERVE) * FONT_MAP_SIZE;
	col = p->ofs_x / FONT_MAP_SIZE;
	row = p->ofs_y / FONT_MAP_SIZE;
	ofs_x = col * FONT_MAP_SIZE - p->ofs_x;
	ofs_y = row * FONT_MAP_SIZE - p->ofs_y;
	cmax = (ofs_x == 0)? FONT_MAP_WIDTH : FONT_MAP_WIDTH  + 1;
	rmax = (ofs_y == 0)? FONT_MAP_HEIGHT: FONT_MAP_HEIGHT + 1;

	for (ro = 0; ro < rmax; ro++) {
		int r = (FONT_MAP_HEIGHT + FONT_MAP_RESERVE + row + ro) % (FONT_MAP_HEIGHT + FONT_MAP_RESERVE);
		int y = ofs_y + ro * FONT_MAP_SIZE;

		for (co = 0; co < cmax; co++) {
			int c = (FONT_MAP_WIDTH + FONT_MAP_RESERVE + col + co) % (FONT_MAP_WIDTH + FONT_MAP_RESERVE);
			MAP_ATTR *attr = &p->map[r][c];
#if 0
			if (c < 0 || c >= FONT_MAP_WIDTH + FONT_MAP_RESERVE
				|| r < 0 || r >= FONT_MAP_HEIGHT + FONT_MAP_RESERVE) {
				fprintf(stderr, "p->ofs_x=%d, p->ofs_y=%d, c=%d, r=%d, row=%d, ro=%d, col=%d, co=%d\n", p->ofs_x, p->ofs_y, c, r, row, ro, col, co);
			}
#endif
			if (attr->map_id) {
				int x = ofs_x + co * FONT_MAP_SIZE;
#if (FONT_MAP_SIZE == 8)
				surface_draw_frame8(screen, attr->surface_id, attr->map_id, x, y);
#elif (FONT_MAP_SIZE == 16)
				surface_draw_frame16(screen, attr->surface_id, attr->map_id, x, y);
#elif (FONT_MAP_SIZE == 32)
				surface_draw_frame32(screen, attr->surface_id, attr->map_id, x, y);
#else
				surface_draw_frame(screen, attr->surface_id, attr->map_id, x, y);
#endif
			}
		}
	}
}

void font_frame(int x, int y, int w, int h) {

	FONTMAP *p = &fontmap;

	p->frame_x = x;
	p->frame_y = y;
	p->frame_w = w;
	p->frame_h = h;
}

void font_default_frame(void) {

	FONTMAP *p = &fontmap;

	p->frame_x = 1;
	p->frame_y = 1;
	p->frame_w = FONT_MAP_WIDTH;
	p->frame_h = FONT_MAP_HEIGHT;
}

void font_attribute(int use) {

	FONTMAP *p = &fontmap;

	if (use)
		p->options |= OPT_USE_ATTRIBUTE;
	else
		p->options &= ~OPT_USE_ATTRIBUTE;
}

void font_locate(int x, int y) {

	FONTMAP *p = &fontmap;

	p->font_x = x;
	p->font_y = y;
}

void font_printf(const char *format, ...) {

	const static font_tex_ids[] = {
		TEX_FONT, TEX_FONT_RED, TEX_FONT_GREEN, TEX_FONT_BLUE, TEX_FONT_YELLOW
	};
	static char buf[FONT_MAP_WIDTH * FONT_MAP_HEIGHT];

	FONTMAP *p = &fontmap;
	MAP_ATTR *attr;

	char *str = buf;
	unsigned char ch;

	va_list ap;

	va_start(ap, format);

	vsnprintf(buf, sizeof(buf), format, ap);

	va_end(ap);

	if (p->options & OPT_USE_ATTRIBUTE) {
		while (p->font_y < p->frame_h && (ch = *str++) != '\0') {

			//printf("font_x=%d, font_y=%d, frame_x=%d, frame_y=%d\n", p->font_x, p->font_y, p->frame_x, p->frame_y);

			switch (ch) {
/*
			case ' ':
				attr = &p->map[p->font_y + p->frame_y][p->font_x + p->frame_x];
				attr->surface_id = fontmap.tex_id;
				attr->map_id = 0;
				p->font_x++;
				break;
*/
			case '\1': /* default */
			case '\2': /* red     */
			case '\3': /* green   */
			case '\4': /* blue    */
			case '\5': /* yellow  */
				font_texture(font_tex_ids[ch - 1]);
				break;
			case '\t':
				p->font_x += 4;
				break;
			case '\n':
				p->font_x = 0;
				p->font_y++;
				break;
			default:
				attr = &p->map[p->font_y + p->frame_y][p->font_x + p->frame_x];
				attr->surface_id = fontmap.tex_id;
				attr->map_id = ch;
				p->font_x++;
				break;
			}

			if (p->font_x >= p->frame_w) {
				p->font_x = 0;
				p->font_y++;
			}
		}
	} else {
		/* 𗘗pȂiׂĕR[hƂĈj */
		while (p->font_y < p->frame_h && (ch = *str++) != '\0') {
			attr = &p->map[p->font_y + p->frame_y][p->font_x + p->frame_x];
			attr->surface_id = fontmap.tex_id;
			attr->map_id = ch;
			p->font_x++;

			if (p->font_x >= p->frame_w) {
				p->font_x = 0;
				p->font_y++;
			}
		}
	}
}
