/*
 * main.c
 *
 * Last modified: <2009/11/28 22:12:10 +0900> By SUMIDA
 */
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <SDL.h>
#include <SDL_image.h>
#include "system.h"
#include "texture.h"
#include "surface.h"
#include "sprite.h"
#include "bg.h"
#include "input.h"
#include "ship.h"
#include "shot.h"
#include "enemy.h"
#include "effect.h"
#include "map.h"
#include "sound.h"
#include "myship.h"
#include "rankin.h"

#define TITLE_TIME   (40*60)
#define RANKING_TIME (20*60)
#define OVER_TIME    (15*60)

const char *datname = "StarFight.dat";
const char *env_appdata = "APPDATA";
char datpath[_MAX_PATH];

SDL_Surface *screen = NULL;

SYSTEM_INFO system_info;

SHIP_MGR  ship_mgr;
SHIP_MGR  enemy_mgr;
SHOT_MGR  ship_shot_mgr;
SHOT_MGR  enemy_shot_mgr;

SHIP_INFO ships[SHIP_MAX];
SHIP_INFO enemy[ENEMY_MAX];
SHOT_INFO ship_shot[SHIP_SHOT_MAX];
SHOT_INFO enemy_shot[ENEMY_SHOT_MAX];

unsigned int frame_count = 0;

PLAYER_INFO player_info;

typedef struct LEVEL_INFO {
	Uint32 opt;
	char *name;
} LEVEL_INFO;

LEVEL_INFO level_infos[] = {
	{SYSTEM_OPT_EASY  , "EASY"  },
	{SYSTEM_OPT_NORMAL, "NORMAL"},
	{SYSTEM_OPT_HARD  , "HARD"  },
	{SYSTEM_OPT_SUPER , "SUPER" },
	{SYSTEM_OPT_HYPER , "HYPER" },
};
int select_level = 1;

static void background_init(void);
static void system_info_draw(SDL_Surface* screen);
static void restart(void);

/* B
 * Ƃ0AsƂ-1ԂB
 */
int Initialize(void) {

	/* [APPDATA]ϐ擾 */
	char *appdata = getenv(env_appdata);
	int i;

	printf("Initialize ... start.\n");

	/* f[^t@C̃pX𐶐 */
	if (appdata == NULL) {
		snprintf(datpath, sizeof(datpath), "%s", datname);
	} else {
		snprintf(datpath, sizeof(datpath), "%s\\%s", appdata, datname);
	}
	printf("DATPATH=%s\n", datpath);

	/*  */
	srand(time(NULL));

	/* SDL */
	if (SDL_Init(SDL_INIT_VIDEO
#ifdef USE_AUDIO
				 |SDL_INIT_AUDIO
#endif
				 |SDL_INIT_TIMER|SDL_INIT_JOYSTICK) < 0) {
		fprintf(stderr, "SDL̏Ɏs܂F%s\n", SDL_GetError());
		return -1;
	}
	printf("SDL_Init ... done.\n");

	/* ʂ̃^CgύX */
	printf("SDL_WM_SetCaption ... ");
	SDL_WM_SetCaption("Star Fight", NULL);
	printf("done.\n");

	/* ʂ */
	printf("SDL_SetVideoMode ... ");
	screen = SDL_SetVideoMode(WIDTH, HEIGHT, DEPTH,
							  (system_info.options & SYSTEM_OPT_FULLSCREEN) ?
							  SDL_FULLSCREEN|SDL_HWSURFACE|SDL_DOUBLEBUF:
							  SDL_SWSURFACE);
	if (screen == NULL) {
		fprintf(stderr, "ʂ̏Ɏs܂F%s\n", SDL_GetError());
		SDL_Quit();
		return -1;
	}
	printf("done.\n");

#ifdef USE_AUDIO
	/*
	 * I[fBI
	 * open 44.1KHz, signed 16bit, system byte order,
	 *      stereo audio, using 4096 byte chunks
	 */
	printf("Mix_OpenAudio ... ");
	if(Mix_OpenAudio(44100, MIX_DEFAULT_FORMAT, 2, 4096) < 0) {
		fprintf(stderr, "I[fBȈɎs܂F%s\n", SDL_GetError());
		SDL_Quit();
		return -1;
	}
	printf("done.\n");
#endif

	if (system_info.options & SYSTEM_OPT_FULLSCREEN) {
		printf("SDL_ShowCursor ... ");
		SDL_ShowCursor(SDL_DISABLE);
		printf("done.\n");
	}

	input_init();
	surface_init();
	sprite_init();
	bg_init();
	font_init();
	effect_init();
	sound_init();

	/* eNX`֘At@C̐ݒ */
	surface_load("image/font16x16.png", TEX_FONT);
	surface_set_defckey(TEX_FONT);

	surface_load("image/font16x16_red.png", TEX_FONT_RED);
	surface_load("image/font16x16_blue.png", TEX_FONT_BLUE);
	surface_load("image/font16x16_green.png", TEX_FONT_GREEN);
	surface_load("image/font16x16_yellow.png", TEX_FONT_YELLOW);

	surface_load("image/bg.png", TEX_BG1);
	surface_set_defckey(TEX_BG1);

	surface_load("image/title.png", TEX_TITLE);
	surface_set_defckey(TEX_TITLE);

	surface_load("image/gameover.png", TEX_GAMEOVER);
	surface_set_defckey(TEX_GAMEOVER);

	surface_load("image/pause.png", TEX_PAUSE);
	surface_set_defckey(TEX_PAUSE);

	surface_load("image/explosion.png", TEX_EXPLOSION);
	surface_set_defckey(TEX_EXPLOSION);
	surface_set_frame(TEX_EXPLOSION, 64, 64);

	surface_load("image/explosion_s.png", TEX_EXPLOSION_S);
	surface_set_defckey(TEX_EXPLOSION_S);
	surface_set_frame(TEX_EXPLOSION_S, 32, 32);

	surface_load("image/bomb.png", TEX_BOMB);
	surface_set_defckey(TEX_BOMB);
	surface_set_frame(TEX_BOMB, 240, 240);

	surface_load("image/fire.png", TEX_FIRE);
	surface_set_defckey(TEX_FIRE);
	surface_set_frame(TEX_FIRE, 56, 56);

	surface_load("image/ship.png", TEX_SHIP);
	surface_set_frame(TEX_SHIP, 60, 60);
	surface_load("image/hit.png", TEX_HIT);
	surface_set_defckey(TEX_HIT);
	surface_set_frame(TEX_HIT, 32, 32);

	surface_load("image/search.png", TEX_SEARCH);
	surface_set_defckey(TEX_SEARCH);
	surface_set_frame(TEX_SEARCH, 64, 64);
	surface_set_alpha(TEX_SEARCH, 200);

	surface_load("image/enemy1.png", TEX_ENEMY1);
	surface_set_frame(TEX_ENEMY1, 60, 60);
	surface_load("image/enemy2.png", TEX_ENEMY2);
	surface_set_frame(TEX_ENEMY2, 60, 60);
	surface_load("image/enemy3.png", TEX_ENEMY3);
	surface_set_frame(TEX_ENEMY3, 60, 60);
	surface_load("image/enemy4.png", TEX_ENEMY4);
	surface_set_frame(TEX_ENEMY4, 60, 60);
	surface_load("image/enemy5.png", TEX_ENEMY5);
	surface_set_frame(TEX_ENEMY5, 60, 60);

	surface_load("image/missile.png", TEX_MISSILE);
	surface_set_defckey(TEX_MISSILE);
	surface_set_frame(TEX_MISSILE, 16, 45);

	surface_load("image/rock.png", TEX_ROCK);
	surface_set_defckey(TEX_ROCK);
	surface_load("image/rockm1.png", TEX_ROCK_M1);
	surface_set_defckey(TEX_ROCK_M1);
	surface_load("image/rockm2.png", TEX_ROCK_M2);
	surface_set_defckey(TEX_ROCK_M2);
	surface_load("image/rockm3.png", TEX_ROCK_M3);
	surface_set_defckey(TEX_ROCK_M3);
	surface_load("image/rocks1.png", TEX_ROCK_S1);
	surface_set_defckey(TEX_ROCK_S1);
	surface_load("image/rocks2.png", TEX_ROCK_S2);
	surface_set_defckey(TEX_ROCK_S2);
	surface_load("image/rocks3.png", TEX_ROCK_S3);
	surface_set_defckey(TEX_ROCK_S3);

	surface_load("image/boss1.png", TEX_BOSS1);
	surface_set_frame(TEX_BOSS1, 100, 181);
	surface_load("image/boss2.png", TEX_BOSS2);
	surface_set_frame(TEX_BOSS2, 100, 150);

	surface_load("image/shot.png", TEX_SHOT);
	surface_set_defckey(TEX_SHOT);
	surface_set_frame(TEX_SHOT, 28, 16);
	surface_load("image/ballet.png", TEX_BALLET);
	surface_set_defckey(TEX_BALLET);
	surface_set_frame(TEX_BALLET, 8, 8);
	surface_load("image/mballet.png", TEX_MBALLET);
	surface_set_defckey(TEX_MBALLET);
	surface_set_frame(TEX_MBALLET, 8, 8);
	surface_load("image/ray1.png", TEX_RAY1);
	surface_set_defckey(TEX_RAY1);
	surface_load("image/ray2.png", TEX_RAY2);
	surface_set_defckey(TEX_RAY2);
	surface_load("image/ray3.png", TEX_RAY3);
	surface_set_defckey(TEX_RAY3);
	surface_load("image/photon.png", TEX_PHOTON);
	surface_set_frame(TEX_PHOTON, 24, 24);
	surface_load("image/photon_s.bmp", TEX_PHOTON_S);
	surface_set_frame(TEX_PHOTON_S, 8, 8);
	//surface_set_alpha(TEX_PHOTON_S, 128);

	surface_load("image/tuki_p1.png", TEX_MOON);
	surface_load("image/wakusei2_p1.png", TEX_PLANET);

	surface_load("image/firesword.png", TEX_ST_FIRE_SWORD);
	surface_load("image/searchlaser.png", TEX_ST_SEARCH_LASER);
	surface_load("image/lockonlaser.png", TEX_ST_LOCKON_LASER);
	surface_load("image/hominglaser.png", TEX_ST_HOMING_LASER);

	surface_load("image/warning_left.png", TEX_WARN_L);
	surface_set_frame(TEX_WARN_L, 102, 16);
	surface_load("image/warning_right.png", TEX_WARN_R);
	surface_set_frame(TEX_WARN_R, 94, 16);

	/* y֘At@C̐ݒ */
	sound_load_music("sound/track010_scramble_ver2.mp3", BGM_SCRAMBLE);
	sound_load_music("sound/Loop_07.mp3" , BGM_01);
	sound_load_music("sound/Loop_113.mp3", BGM_02);
	sound_load_music("sound/Loop_26.mp3" , BGM_03);
	sound_load_music("sound/message_n.mp3", BGM_ENTRY);
	sound_load_effect("sound/tm2_shoot000.wav", SE_SHOT0);
	sound_load_effect("sound/tm2_shoot001.wav", SE_SHOT1);
	sound_load_effect("sound/tm2_sonic000.wav", SE_SHOT2);
	sound_load_effect("sound/tm2_bom001.wav"  , SE_BOMB0);
	sound_load_effect("sound/tm2r_bom26_c.wav", SE_BOMB1);
	sound_load_effect("sound/tm2r_bom31_b.wav", SE_BOMB2);

	/* e}l[W̏ */
	ship_init(&ship_mgr, ships, NUMBEROF(ships));
	ship_init(&enemy_mgr, enemy, NUMBEROF(enemy));

	shot_init(&ship_shot_mgr, ship_shot, NUMBEROF(ship_shot));
	shot_init(&enemy_shot_mgr, enemy_shot, NUMBEROF(enemy_shot));

	system_info.status |= SYSTEM_MODE_READY;

	/* @̐ݒ */
	myship_init(&ship_mgr);
	ship->stat = SHIP_STAT_LOST;

	/* XRAǂݍ */
	score_load();

	/* }bv̐ݒ */
	map_init();

	/* wi̐ݒ */
	background_init();

	/* BGM̍Đ */
	sound_music(BGM_SCRAMBLE, -1);

	printf("Initialize ... done.\n");

	return 0;
}

/* XV */
void Update(void) {

	Uint32 mode = system_info.status & SYSTEM_MODE_MASK;

	/* ͑u̓ǂݍ */
	input_read();

	if (input[P1] & INPUT_KEYDOWN_START) {

		switch (mode) {
		case SYSTEM_MODE_READY:
		case SYSTEM_MODE_RANKING:
			/* IʂɑJ */
			system_info.status &= ~SYSTEM_MODE_MASK;
			system_info.status |= SYSTEM_MODE_SELECT;
			break;

		case SYSTEM_MODE_SELECT:
			/* Q[Jn */
			restart();
			break;

		case SYSTEM_MODE_OVER:
			/* l[Gg[ */
			system_info.count = OVER_TIME;
			break;

		case SYSTEM_MODE_ENTRY:
			break;

		case SYSTEM_MODE_SCORE:
			/* Q[ĊJ */
			system_info.status &= ~SYSTEM_MODE_MASK;
			system_info.status |= SYSTEM_MODE_READY;
			system_info.count = 0;
			sound_music(BGM_SCRAMBLE, -1);
			break;

		case SYSTEM_MODE_PAUSE:
			/* |[Y@\̉ */
			system_info.status &= ~SYSTEM_MODE_PAUSE;
			break;

		default:
			/* |[Y@\̐ݒ */
			system_info.status |= SYSTEM_MODE_PAUSE;
			break;
		}
	}

	mode = system_info.status & SYSTEM_MODE_MASK;

	switch (mode) {
	case SYSTEM_MODE_SELECT:
		if (input[P1] & (INPUT_KEYDOWN_BUTTON1
						 | INPUT_KEYDOWN_RIGHT | INPUT_KEYDOWN_DOWN)) {
			laser_type = (laser_type + 1) % LASER_TYPE_MAX;
		}
		if (input[P1] & (INPUT_KEYDOWN_BUTTON2
						 | INPUT_KEYDOWN_LEFT | INPUT_KEYDOWN_UP)) {
			laser_type = (laser_type - 1 + LASER_TYPE_MAX) % LASER_TYPE_MAX;
		}
		break;

	case SYSTEM_MODE_READY:
		if (++system_info.count > TITLE_TIME) {
			rank = -1;
			system_info.count = 0;
			system_info.status &= ~SYSTEM_MODE_MASK;
			system_info.status |= SYSTEM_MODE_RANKING;
		} else {

			if ((input[P1] & INPUT_KEYDOWN_RIGHT)
				&& select_level < NUMBEROF(level_infos) - 1) {
				++select_level;
			}
			if ((input[P1] & INPUT_KEYDOWN_LEFT)
				&& select_level > 0) {
				--select_level;
			}
			system_info.options &= ~SYSTEM_OPT_LEVEL_TYPE;
			system_info.options |= SYSTEM_OPT_LEVEL| level_infos[select_level].opt;
			system_info.level = level_infos[select_level].opt;
		}
		break;

	case SYSTEM_MODE_RANKING:
		if (++system_info.count > RANKING_TIME) {
			system_info.count = 0;
			system_info.status &= ~SYSTEM_MODE_MASK;
			system_info.status |= SYSTEM_MODE_READY;
			sound_music(BGM_SCRAMBLE, -1);
		}
		break;

	case SYSTEM_MODE_OVER:
		if (++system_info.count > OVER_TIME || !sound_playing_music()) {
			system_info.count = 0;
			system_info.status &= ~SYSTEM_MODE_MASK;
			if (rankin_init() >= 0) {
				system_info.status |= SYSTEM_MODE_ENTRY;
				sound_music(BGM_ENTRY, -1);
			} else {
				system_info.status |= SYSTEM_MODE_SCORE;
			}
		}

	case SYSTEM_MODE_ENTRY:
		if (rankin_action()) {
			system_info.count = 0;
			system_info.status &= ~SYSTEM_MODE_MASK;
			system_info.status |= SYSTEM_MODE_SCORE;
		}
		break;

	case SYSTEM_MODE_SCORE:
		if (++system_info.count > RANKING_TIME) {
			system_info.count = 0;
			system_info.status &= ~SYSTEM_MODE_MASK;
			system_info.status |= SYSTEM_MODE_READY;
			sound_music(BGM_SCRAMBLE, -1);
		}
		break;

	default:
		break;
	}

	if (mode != SYSTEM_MODE_PAUSE) {

		/* \XvCgj */
		sprite_clean();

		/* wĩXN[ */
		bg[1].ofs_y -= 3;
		bg[2].ofs_y -= 1;
		/*bg[3].ofs_y -= 1;/**/

		/* }bv̐ */
		if (!(system_info.status & SYSTEM_MODE_MASK)) {
			map_action();
		}

		/* @Vbg̐ */
		system_info.score += shot_action(&ship_shot_mgr, &enemy_mgr);

		/* G@Vbg̐ */
		system_info.score += shot_action(&enemy_shot_mgr, &ship_mgr);

		/* @̐ */
		myship_action(&ship_mgr, &ship_shot_mgr, &enemy_mgr);

		/* G@̐ */
		enemy_action(&enemy_mgr, &enemy_shot_mgr, ship);

		/* ʌʂ̐ */
		effect_action();

		/* XRA\ */
		font_locate(0, 0);
		font_default_texture();
		font_printf("\3SCORE:%08d  LEVEL:%02d", system_info.score, system_info.level);
	}
}

/* `悷 */
void Draw(void) {

	/* wi̐F` */
	SDL_FillRect(screen, NULL, 0);

	/* wiƃXvCg`悷 */
	/*bg_rendaring(screen, 3); /**/
	bg_rendaring(screen, 2);
	bg_rendaring(screen, 1);
	sprite_rendering(screen);
	/*bg_rendaring(screen, 0); /**/
	myship_rendaring();
	font_rendaring(screen);
	system_info_draw(screen);

	/* ʂXV */
	SDL_Flip(screen);

	++frame_count;
}

/* C[v */
void MainLoop(void) {

	SDL_Event event;
	double next_frame = SDL_GetTicks();
	double wait = 1000.0 / 60;

	for (;;) {
		/* ׂẴCxg */
		while (SDL_PollEvent(&event)) {
			/* QUIT Cxg邩AESC L[ꂽI */
			if ((event.type == SDL_QUIT) ||
				(event.type == SDL_KEYUP && event.key.keysym.sym == SDLK_ESCAPE))
				return;
		}

		/* 1bԂ60Update悤ɂ */
		if (SDL_GetTicks() >= next_frame) {
			Update();

			/* Ԃ܂ƂDraw */
			// pɂɃt[Ԃ̂ŃRgAEg
			//if (SDL_GetTicks() < next_frame + wait)
				Draw();
			next_frame += wait;
			SDL_Delay(0);
		}
	}
}


/* Is */
void Finalize(void) {

	printf("Finalize ... start.\n");

	surface_free();
	input_free();
	sound_free();

	/* XRA̕ۑ */
	score_save();

	/* I */
	SDL_Quit();

	printf("Finalize ... done.\n");
}


/* C֐ */
int SDL_main(int argc, char* argv[]) {

	typedef struct OPTION_INFO {
		char*  command;
		Uint32 value;
	} OPTION_INFO;

	const static OPTION_INFO options[] = {
		{"-v"      , SYSTEM_OPT_VERBOSE},
		{"-f"      , SYSTEM_OPT_FULLSCREEN},
		{"--full"  , SYSTEM_OPT_FULLSCREEN},
	};

	memset(&system_info, 0, sizeof(system_info));
	score_init();

	if (argc > 1) {
		int i;

		for (i = 1; i < argc; i++) {
			char *cmd = argv[i];
			int j;

			if (!strcmp("--help", cmd)) {
				printf("Usage: %s [options]\n"
					   "Options: -v        Verbose mode\n"
					   "         -f or --full Full screen mode\n"
					   );
				return EXIT_SUCCESS;
			}

			for (j = NUMBEROF(options) - 1; j >= 0 ; j--) {
				const OPTION_INFO *opt = &options[j];

				if (!strcmp(opt->command, cmd)) {
					system_info.options |= opt->value;
					break;
				}
			}

			if (j < 0) {
				fprintf(stderr, "Illegal option: %s\n", cmd);
				return EXIT_FAILURE;
			}
		}
	}

	if (Initialize() < 0)
		return EXIT_FAILURE;

	MainLoop();
	Finalize();

	return EXIT_SUCCESS;
}
/*----------------------------------------------------------------------------*/
/* wȉ */
void background_init(void) {

	int x, y;

	for (y = 0; y < BG_MAP_HEIGHT + BG_MAP_RESERVE; y++) {
		for (x = 0; x < BG_MAP_WIDTH + BG_MAP_RESERVE; x++) {
			bg[1].map[y][x].surface_id = TEX_BG1;
			bg[1].map[y][x].map_id = (int)Random(3, 9);
			bg[2].map[y][x].surface_id = TEX_BG1;
			bg[2].map[y][x].map_id = (int)Random(2, 6);
		}
	}
}

/* VXe̕` */
void system_info_draw(SDL_Surface* screen) {

	const static int info_id[] = {-1, TEX_TITLE, -1, -1, TEX_GAMEOVER, -1, -1, TEX_PAUSE};
	const static int spwt_id[] = {TEX_ST_FIRE_SWORD, TEX_ST_SEARCH_LASER, TEX_ST_LOCKON_LASER, TEX_ST_HOMING_LASER};
	const static char *spw_name[] = {"  \3FIRE SWORD", "\3SEARCH LASER", "\3LOCKON LASER", "\3HOMING LASER"};
	const static char *spw_info[] = {
		"\3 Press \5`Z'\3 key, the gauge is collected.\nWhen the key is separated,             \n Fire Sword is shot.                   \n                                       ",
		"\3 Press \5`X'\3 key, Radar extends.         \nCatch the enemy plane.                 \n The search laser is shot at the enemy \nplane in radar.                        ",
		"\3 Press \5`X'\3 key, Radar extends.         \nCatch the enemy plane.                 \n When the key is separated, the lockon \nlaser is shot.                         ",
		"\3 Press \5`X'\3 key, Radar extends.         \nCatch the enemy plane.                 \n When the key is separated, the homing \nlaser is shot.                         ",
	};
	Uint32 mode = system_info.status & SYSTEM_MODE_MASK;
	int id = info_id[mode];

	if (id >= 0) {
		SURFACE *surface = surface_get(id);
		surface_draw(screen, id, (WIDTH - surface->frame_width) / 2, (HEIGHT - surface->frame_height) / 2);
	}

	switch (mode) {
	case SYSTEM_MODE_READY:
		font_clean();
		font_locate(14, 25);
		font_printf("Level:%s", level_infos[select_level].name);
		break;

	case SYSTEM_MODE_SELECT:
		{
			int texid = spwt_id[laser_type];
			SURFACE *surface = surface_get(texid);
			surface_draw(screen, texid, (WIDTH - surface->frame_width) / 2, (HEIGHT - surface->frame_height) / 2);

			font_clean();
			font_locate(0, 4);
			font_printf("\4   Press cursor key to select          \n");
			font_printf(  "  `Special Weapon System'.             ");

			font_locate(0, 20);
			font_printf(spw_info[laser_type]);

			font_locate(8, 25);
			font_printf("\4Press `S' key to Start!\1");
			break;
		}

	case SYSTEM_MODE_RANKING:
	case SYSTEM_MODE_SCORE:
		score_draw();
		break;

	case SYSTEM_MODE_ENTRY:
		rankin_draw();
		break;
	}

	font_locate(27, 28);
	font_default_texture();
	font_printf("%s", spw_name[laser_type]);
}

/* ăX^[g */
void restart(void) {
	system_info.status = 0;
	system_info.score  = 0;
	system_info.frame  = 0;

	ship_init(&ship_mgr, ships, NUMBEROF(ships));
	ship_init(&enemy_mgr, enemy, NUMBEROF(enemy));
	shot_init(&ship_shot_mgr, ship_shot, NUMBEROF(ship_shot));
	shot_init(&enemy_shot_mgr, enemy_shot, NUMBEROF(enemy_shot));

	myship_init(&ship_mgr);
	font_init();
	map_init();
	effect_clean();

	player_info.score = 0;
	player_info.laser_type = laser_type;
	player_info.start_level =
		player_info.finish_level = system_info.level;
	snprintf(player_info.name, sizeof(player_info.name), "%10c", ' ');

	sound_music(BGM_01, -1);
}
/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
/* lower ȏ upper ȉ̗Ԃ */
double Random(double lower, double upper) {
	return ((double)rand() / RAND_MAX) * (upper - lower) + lower;
}

/* Pʉ~̍Wpx߂ */
double angle(double x, double y) {
	if (y == 0.0 && x == 0.0) {
		return 0;

	} else {
		double r = atan2(-y, x) * 180 / M_PI;

		if (r >= -90.0)
			r = r + 90.0;
		else
			r = r + 450.0;
		return r;
	}
}

char *format(const char *format, ...) {

	static char buf[1024];

	va_list ap;
	va_start(ap, format);
	vsnprintf(buf, sizeof(buf), format, ap);
	va_end(ap);

	return buf;
}

void report(const char *file, const int line, const char *message) {

	if (system_info.options & SYSTEM_OPT_VERBOSE) {
		printf("%s(%d) %s\n", file, line, message);
	}
}
