/*
 * map.c
 *
 * Last modified: <2009/11/28 02:29:38 +0900> By SUMIDA
 */
#include <SDL.h>
#include "system.h"
#include "real.h"
#include "surface.h"
#include "sprite.h"
#include "ship.h"
#include "shot.h"
#include "enemy.h"
#include "map.h"
#include "input.h"
#include "sound.h"
#include "effect.h"
#include "texture.h"

#define STAT_READY     0x80
#define STAT_EMPTY     0x10

#define STAT_TASKS     0x0F
#define TASK_BGM1      0x01
#define TASK_BGM2      0x02
#define TASK_MOON      0x03
#define TASK_PLANET    0x04

enum {
	APR_NONE = 0,
	APR_DOWN,
	APR_UP,
	APR_LEFT,
	APR_RIGHT,
	APR_SIDE,
	APR_HOLD,
	APR_ROUND,
	APR_MAX
};

typedef struct MAP_INFO {
	Sint8 id;       /* CxgID   */
	Sint8 count;    /* o       */
	Sint8 interval; /* oԊu     */
	Sint8 apper;    /* op^[ */
	Uint8 options;  /* tOIvV */
	Uint8 status;   /* }bv   */
} MAP_INFO;

/* G@̏o߂B */
static MAP_INFO map[] = {
	{  -1, 1,    0, 0, 0, STAT_EMPTY|TASK_BGM1},
	{  -1, 1,    0, 0, 0, STAT_EMPTY|TASK_MOON},
	{  -1, 1, 60*2, 0, 0, STAT_EMPTY},

#if 0
	/*	{SHIP_ID_ENEMY1, 50, 20, APR_DOWN, 0x00, 0}, /**/
/*	{SHIP_ID_ENEMY1,  5, 20, APR_UP  , 0x00, 0}, /**/
/*	{SHIP_ID_ENEMY1,  4, 20, APR_LEFT, 0x00, 0}, /**/
/*	{SHIP_ID_ENEMY1,  5, 20, APR_RIGHT,0x00, 0}, /**/
/*	{SHIP_ID_ENEMY1,  5, 20, APR_SIDE, 0x00, 0}, /**/
/*	{SHIP_ID_ENEMY1, 50, 20, APR_HOLD, 0x00, 0}, /**/
{SHIP_ID_ENEMY1, 50, 20, APR_ROUND,0x00, 0}, /**/
/*	{SHIP_ID_ENEMY1, 10, 20, APR_DOWN, 0x01, 0}, /**/
/*	{SHIP_ID_ENEMY1,  3, 20, APR_UP  , 0x01, 0}, /**/
/*	{SHIP_ID_ENEMY1,  3, 20, APR_LEFT, 0x01, 0}, /**/
/*	{SHIP_ID_ENEMY1,  3, 20, APR_RIGHT,0x01, 0}, /**/
/*	{SHIP_ID_ENEMY1,  2, 20, APR_SIDE, 0x01, 0}, /**/
/*	{SHIP_ID_ENEMY1,  2, 20, APR_HOLD, 0x01, 0}, /**/
/*	{SHIP_ID_ENEMY1, 10, 20, APR_ROUND,0x01, 0}, /**/
#endif

	{SHIP_ID_ENEMY1, 10, 20, APR_DOWN, 0x00, 0}, /**/
	{SHIP_ID_ENEMY1,  7, 20, APR_LEFT, 0x00, 0}, /**/
	{SHIP_ID_ENEMY1,  7, 20, APR_RIGHT,0x00, 0}, /**/
	{SHIP_ID_ENEMY2, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY3, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY1, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ROCK ,   5, 30, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY1,  1, 20, APR_SIDE, 0x00, 0}, /**/
	{SHIP_ID_ENEMY2, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY3, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY5, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY1,  1, 20, APR_HOLD, 0x00, 0}, /**/
	{SHIP_ID_ENEMY1, 10, 20, APR_ROUND,0x00, 0}, /**/
	{SHIP_ID_ENEMY2, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY1, 10, 20, APR_DOWN, 0x01, 0}, /**/
	{SHIP_ID_ENEMY3, 10, 20, APR_DOWN, 0, 0},
	{  -1, 1,   60, 0, 0, STAT_EMPTY|TASK_BGM2},
	{SHIP_ID_BOSS1 ,  1, 20, 0, 0, 0},
	{  -1, 1,    0, 0, 0, STAT_EMPTY|TASK_BGM1},
	{  -1, 1, 60*2, 0, 0, STAT_EMPTY},
	{SHIP_ID_ROCK ,   5, 30, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY1,  3, 20, APR_UP  , 0x00, 0}, /**/
	{SHIP_ID_ENEMY5, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY4, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY1,  3, 20, APR_LEFT, 0x01, 0}, /**/
	{SHIP_ID_ENEMY1,  3, 20, APR_RIGHT,0x01, 0}, /**/
	{SHIP_ID_ENEMY3, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY1, 10, 20, APR_DOWN, 0, 0},
	{  -1, 1,    0, 0, 0, STAT_EMPTY|TASK_PLANET},
	{SHIP_ID_ROCK ,   5, 30, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY2, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY4, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY1,  2, 20, APR_SIDE, 0x01, 0}, /**/
	{SHIP_ID_ENEMY3, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY1, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY5, 10, 20, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY1,  1, 20, APR_HOLD, 0x01, 0}, /**/
	{SHIP_ID_ROCK ,   5, 30, APR_DOWN, 0, 0},
	{SHIP_ID_ENEMY1,  1, 20, APR_ROUND,0x01, 0}, /**/
	{  -1, 1,   60, 0, 0, STAT_EMPTY|TASK_BGM2},
	{SHIP_ID_BOSS2 ,  1, 20, 0, 0, 0}
};

extern SHIP_MGR enemy_mgr;

static MAP_INFO curr_map;
static void map_generate_enemy(void);
static void map_action_warning(SPRITE *sprite);

static void map_action_bgm1(void);
static void map_action_bgm2(void);
static void map_action_moon(void);
static void map_action_planet(void);

static void(*func_action[])(void) = {
	NULL,
	map_action_bgm1,
	map_action_bgm2,
	map_action_moon,
	map_action_planet,
};

void map_init(void) {
	system_info.level = (system_info.options & SYSTEM_OPT_LEVEL)?
		system_info.options & SYSTEM_OPT_LEVEL_TYPE: SYSTEM_OPT_EASY;
	//		system_info.options & SYSTEM_OPT_LEVEL_TYPE: SYSTEM_OPT_NORNAL;
	//		system_info.options & SYSTEM_OPT_LEVEL_TYPE: SYSTEM_OPT_HYPER;
	system_info.count     = 0;
	system_info.encounter = 0;
	system_info.map_index = 0;
	memset(&curr_map, 0, sizeof(curr_map));
}

void map_action(void) {

	if (system_info.status & SYSTEM_STAT_BOSS) return;

	if (system_info.encounter > 0) {
		--system_info.encounter;
		return;
	}

	if (!(curr_map.status & STAT_READY)) {
		memcpy(&curr_map, &map[system_info.map_index], sizeof(curr_map));
		curr_map.status |= STAT_READY;
	}

#if 0
	if (input[P1] & INPUT_KEYUP_SELECT) {
		//SHIP_INFO *enemy = enemy_new(&enemy_mgr, SHIP_ID_BOSS1, 0, 0);
		SHIP_INFO *enemy = enemy_new(&enemy_mgr, SHIP_ID_BOSS2, 0, 0);
		//SHIP_INFO *enemy = enemy_new(&enemy_mgr, SHIP_ID_ENEMY1, 0, 0);
		//SHIP_INFO *enemy = enemy_new(&enemy_mgr, SHIP_ID_ROCK, 0, 0);
		//SHIP_INFO *enemy = enemy_new(&enemy_mgr, (int)Random(SHIP_ID_ENEMY1, SHIP_ID_MAX), 0, 0);

		if (enemy != NULL) {
			SURFACE *surface = surface_get(enemy->sprite.surface_id);

			enemy->sprite.x = IntToReal((int)Random(0, WIDTH - surface->frame_width));
			enemy->sprite.y = IntToReal(-surface->frame_height);
			enemy->sprite.dx = IntToReal((int)Random(-2, 2));
		}
	}
#endif

#if 1
	if (--curr_map.interval < 0) {
		if (!(curr_map.status & STAT_EMPTY)) {
			/* G@̐ */
			map_generate_enemy();

		} else {
			int task = curr_map.status & STAT_TASKS;

			if (task < NUMBEROF(func_action) && func_action[task] != NULL) {
				func_action[task]();
			}

			curr_map.status &= ~STAT_TASKS;
		}

		if (--curr_map.count > 0) {
			curr_map.interval = map[system_info.map_index].interval;
		} else {
			system_info.encounter = 60;
			curr_map.status &= ~STAT_READY;
			if (++system_info.map_index >= NUMBEROF(map)) {
				system_info.map_index = 0;
				if (++system_info.level > LEVEL_MAX) system_info.level = LEVEL_MAX;
			}
		}
	}
#endif
}

void map_generate_enemy(void) {

	/*
	SHIP_INFO *enemy = enemy_new(&enemy_mgr, curr_map.id, 0, 0);

	if (enemy != NULL) {
		SURFACE *surface = surface_get(enemy->sprite.surface_id);

		enemy->flag |= curr_map.options << 16;
		enemy->sprite.x = IntToReal((int)Random(0, WIDTH / surface->frame_width) * surface->frame_width);
		enemy->sprite.y = IntToReal(-surface->frame_height);
		enemy->sprite.dx = IntToReal((int)Random(-2, 2));
	}
	 */

	switch (curr_map.apper) {
	default:
	case APR_DOWN:
		{
			SHIP_INFO *enemy = enemy_new(&enemy_mgr, curr_map.id, 0, 0);

			if (enemy != NULL) {
				SURFACE *surface = surface_get(enemy->sprite.surface_id);
				int cnt = WIDTH / surface->frame_width;
				int pos = (int)Random(0, cnt);

				enemy->flag |= curr_map.options << 16;
				enemy->sprite.x = IntToReal(pos * surface->frame_width);
				enemy->sprite.y = IntToReal(-surface->frame_height);
				enemy->sprite.dx = IntToReal((int)(Random(0, 1) * (cnt / 2 - pos)/2));
			}
		}
		break;

	case APR_UP:
		{
			SHIP_INFO *enemy = enemy_new(&enemy_mgr, curr_map.id, 0, 0);

			if (enemy != NULL) {
				SURFACE *surface = surface_get(enemy->sprite.surface_id);

				enemy->flag |= curr_map.options << 16;
				enemy->sprite.x = IntToReal((int)Random(0, WIDTH / surface->frame_width) * surface->frame_width);
				enemy->sprite.y = IntToReal(HEIGHT);
				enemy->sprite.dy = IntToReal(-1);

				map_action_warning(&enemy->sprite);
			}
		}
		break;

	case APR_LEFT:
		{
			SHIP_INFO *enemy = enemy_new(&enemy_mgr, curr_map.id, 0, 0);

			if (enemy != NULL) {
				SURFACE *surface = surface_get(enemy->sprite.surface_id);

				enemy->flag |= curr_map.options << 16;
				enemy->sprite.x = IntToReal(-surface->frame_width);
				//enemy->sprite.y = IntToReal((int)Random(
				//	surface->frame_height * 4,
				//	HEIGHT / surface->frame_height) * surface->frame_height);
				enemy->sprite.y = IntToReal(
					HEIGHT / 2 - surface->frame_height * (curr_map.count % 5));
				enemy->sprite.dx = IntToReal(2);
				enemy->sprite.dy = IntToReal(1);

				map_action_warning(&enemy->sprite);
			}
		}
		break;

	case APR_RIGHT:
		{
			SHIP_INFO *enemy = enemy_new(&enemy_mgr, curr_map.id, 0, 0);

			if (enemy != NULL) {
				SURFACE *surface = surface_get(enemy->sprite.surface_id);

				enemy->flag |= curr_map.options << 16;
				enemy->sprite.x = IntToReal(WIDTH);
				//enemy->sprite.y = IntToReal((int)Random(
				//	surface->frame_height * 4,
				//	HEIGHT / surface->frame_height) * surface->frame_height);
				enemy->sprite.y = IntToReal(
					HEIGHT / 2 - surface->frame_height * (curr_map.count % 5));
				enemy->sprite.dx = IntToReal(-2);
				enemy->sprite.dy = IntToReal(1);

				map_action_warning(&enemy->sprite);
			}
		}
		break;

	case APR_SIDE:
		{
			int i;

			for (i = 0; i < 4; i++) {
				SHIP_INFO *enemy = enemy_new(&enemy_mgr, curr_map.id, 0, 0);

				if (enemy != NULL) {
					SURFACE *surface = surface_get(enemy->sprite.surface_id);

					enemy->flag |= curr_map.options << 16;
					enemy->sprite.x = IntToReal(-surface->frame_width);
					enemy->sprite.y = IntToReal(surface->frame_height * 2 * i);
					enemy->sprite.dx = IntToReal(2);
					enemy->sprite.dy = IntToReal(0);

					map_action_warning(&enemy->sprite);
				}
			}
		}
		{
			int i;

			for (i = 0; i < 4; i++) {
				SHIP_INFO *enemy = enemy_new(&enemy_mgr, curr_map.id, 0, 0);

				if (enemy != NULL) {
					SURFACE *surface = surface_get(enemy->sprite.surface_id);

					enemy->flag |= curr_map.options << 16;
					enemy->sprite.x = IntToReal(WIDTH);
					enemy->sprite.y = IntToReal(surface->frame_height * 2 * i + surface->frame_height);
					enemy->sprite.dx = IntToReal(-2);
					enemy->sprite.dy = IntToReal(0);

					map_action_warning(&enemy->sprite);
				}
			}
		}
		break;

	case APR_HOLD:
		{
			SHIP_INFO *enemy = enemy_new(&enemy_mgr, curr_map.id, 0, 0);

			if (enemy != NULL) {
				SURFACE *surface = surface_get(enemy->sprite.surface_id);

				enemy->flag |= curr_map.options << 16;
				//enemy->sprite.x = IntToReal((int)Random(0, WIDTH / surface->frame_width) * surface->frame_width);
				enemy->sprite.x = IntToReal(surface->frame_width * 2 * (curr_map.count % 5));
				enemy->sprite.y = IntToReal(-surface->frame_height);
				//enemy->sprite.dx = IntToReal((int)Random(-2, 2));
				enemy->sprite.dx = IntToReal(0);
				enemy->sprite.dy = IntToReal(2);

				map_action_warning(&enemy->sprite);
			}
		}
		{
			SHIP_INFO *enemy = enemy_new(&enemy_mgr, curr_map.id, 0, 0);

			if (enemy != NULL) {
				SURFACE *surface = surface_get(enemy->sprite.surface_id);

				enemy->flag |= curr_map.options << 16;
				//enemy->sprite.x = IntToReal((int)Random(0, WIDTH / surface->frame_width) * surface->frame_width);
				enemy->sprite.x = IntToReal(WIDTH - surface->frame_width * 2 * (curr_map.count % 5));
				enemy->sprite.y = IntToReal(HEIGHT);
				enemy->sprite.dx = IntToReal(0);
				enemy->sprite.dy = IntToReal(-2);

				map_action_warning(&enemy->sprite);
			}
		}
		break;

	case APR_ROUND:
		{
			int c = curr_map.count / 4;
			switch (curr_map.count % 4) {
			default:
			case 0: /*  */
				{
					SHIP_INFO *enemy = enemy_new(&enemy_mgr, curr_map.id, 0, 0);

					if (enemy != NULL) {
						SURFACE *surface = surface_get(enemy->sprite.surface_id);

						enemy->flag |= curr_map.options << 16;
						enemy->sprite.x = IntToReal(
							(WIDTH - surface->frame_width * (c + 1)) % WIDTH);
						enemy->sprite.y = IntToReal(-surface->frame_height);
						//enemy->sprite.dx = IntToReal((int)Random(-2, 2));
						enemy->sprite.dx = IntToReal(0);
						enemy->sprite.dy = IntToReal(2);

						map_action_warning(&enemy->sprite);
					}
				}
				break;

			case 1: /*  */
				{
					SHIP_INFO *enemy = enemy_new(&enemy_mgr, curr_map.id, 0, 0);

					if (enemy != NULL) {
						SURFACE *surface = surface_get(enemy->sprite.surface_id);

						enemy->flag |= curr_map.options << 16;
						enemy->sprite.x = IntToReal(-surface->frame_width);
						enemy->sprite.y = IntToReal(
							(HEIGHT - surface->frame_height * (c + 1)) % HEIGHT);
						//enemy->sprite.dx = IntToReal((int)Random(1, 2));
						enemy->sprite.dx = IntToReal(2);
						enemy->sprite.dy = IntToReal(0);

						map_action_warning(&enemy->sprite);
					}
				}
				break;

			case 2: /*  */
				{
					SHIP_INFO *enemy = enemy_new(&enemy_mgr, curr_map.id, 0, 0);

					if (enemy != NULL) {
						SURFACE *surface = surface_get(enemy->sprite.surface_id);

						enemy->flag |= curr_map.options << 16;
						//enemy->sprite.x = IntToReal(
						//	(int)Random(0, WIDTH / surface->frame_width) * surface->frame_width);
						enemy->sprite.x = IntToReal((surface->frame_width * c) % WIDTH);
						enemy->sprite.y = IntToReal(HEIGHT);
						enemy->sprite.dx = IntToReal(0);
						enemy->sprite.dy = IntToReal(-2);

						map_action_warning(&enemy->sprite);
					}
				}
				break;

			case 3: /*  */
				{
					SHIP_INFO *enemy = enemy_new(&enemy_mgr, curr_map.id, 0, 0);

					if (enemy != NULL) {
						SURFACE *surface = surface_get(enemy->sprite.surface_id);

						enemy->flag |= curr_map.options << 16;
						enemy->sprite.x = IntToReal(WIDTH);
						enemy->sprite.y = IntToReal((surface->frame_height * c) % HEIGHT);
						//enemy->sprite.y = IntToReal(
						//	(int)Random(0, HEIGHT / surface->frame_height) * surface->frame_height);
						//enemy->sprite.dx = IntToReal((int)Random(1, 2));
						enemy->sprite.dx = IntToReal(-2);
						enemy->sprite.dy = IntToReal(0);

						map_action_warning(&enemy->sprite);
					}
				}
				break;
			}
		}
		break;
	}
}

void map_action_bgm1(void) {

	sound_music(BGM_01, -1);
}

void map_action_bgm2(void) {

	sound_music(BGM_02, -1);
}

void map_action_moon(void) {

	EFFECT_INFO *effect = effect_new(EFFECT_ID_MOON);

	if (effect != NULL) {
		SURFACE *surface = surface_get(effect->sprite.surface_id);

		effect->sprite.x = IntToReal((int)Random(0, WIDTH / surface->frame_width) * surface->frame_width);
		effect->sprite.y = IntToReal(-surface->frame_height);
		effect->sprite.dx = 0;
		effect->sprite.dy = 0x1000;
	}
}

void map_action_planet(void) {

	EFFECT_INFO *effect = effect_new(EFFECT_ID_PLANET);

	if (effect != NULL) {
		SURFACE *surface = surface_get(effect->sprite.surface_id);

		effect->sprite.x = IntToReal(
			(int)Random(0, WIDTH / surface->frame_width) * surface->frame_width);
		effect->sprite.y = IntToReal(-surface->frame_height);
		effect->sprite.dx = 0;
		effect->sprite.dy = 0x2000;
	}
}

void map_action_warning(SPRITE *sprite) {

	SURFACE *s0 = surface_get(sprite->surface_id);
	SURFACE *s1 = surface_get(TEX_WARN_L);
	SURFACE *s2 = surface_get(TEX_WARN_R);
	Real x = sprite->x + (IntToReal(s0->frame_width) >> 1);
	Real y = sprite->y;
	Real sy = IntToReal(HEIGHT - s1->frame_height);
	EFFECT_ID id;
	EFFECT_INFO *effect;

	if (x < 0) x = 0;
	if (x >= IntToReal(WIDTH)) x = IntToReal(WIDTH);

	if (x < IntToReal(WIDTH) >> 1) {
		id = EFFECT_ID_WARNING_LEFT;
	} else {
		id = EFFECT_ID_WARNING_RIGHT;
		x -= IntToReal(s2->frame_width);
	}

	if (y <   0) y = 0;
	if (y >= sy) y = sy;

	effect = effect_new(id);
	if (effect != NULL) {
		effect->sprite.x = x;
		effect->sprite.y = y;
	}
}
