/*
 * myship.c
 *
 * Last modified: <2009/11/24 20:37:30 +0900> By SUMIDA
 */
#include <stdlib.h>
#include <math.h>
#include <SDL.h>
#include <SDL_image.h>
#include "system.h"
#include "texture.h"
#include "surface.h"
#include "sprite.h"
#include "input.h"
#include "ship.h"
#include "shot.h"
#include "enemy.h"
#include "myship.h"
#include "effect.h"
#include "sound.h"
#include "draw.h"
#include "bg.h"

#define SIDE_SPACE       30
#define BOTTOM_SPACE     30

#define MYSHIP_SPEED     4

#define SWORD_WAIT       (SHIP_SHOT_MAX * 6)
#define SWORD_POWER_MAX  100

#define FIRE_SWORD       0x0100
#define FIRE_SWORD_W     40

#define RADAR_SPEED      4
#define RADAR_R          200
#define LOCKON_MAX       8

#define STRENGTH_SWORD   2
#define STRENGTH_SEARCH  4
#define STRENGTH_LOCKON  2

#define SEARCH_LASER     0x0200
#define LASER_TIME_MAX   100

typedef struct RANGE_INFO {
	SHIP_INFO *ship;
	int range;
} RANGE_INFO;

extern SHIP_INFO enemy[ENEMY_MAX];
extern SHIP_MGR  enemy_mgr;
extern SHOT_MGR  enemy_shot_mgr;
extern SDL_Surface *screen;
extern unsigned int frame_count;
extern SHIP_TYPE ship_type[];

SHIP_INFO *ship  = NULL;

RANGE_INFO range_infos[ENEMY_MAX];
RANGE_INFO lockon[LOCKON_MAX];
int lockon_cnt = 0;

int sword = 0;           /* \[ho̓Q[W */
int sword_enabled = 0;   /* \[h[UQ[W */

int radar = 0;           /* [_[a     */
int radar_enabled = 0;   /* [_[JE^ */

LASER_TYPE laser_type = LASER_TYPE_SWORD;

/* oJVbg */
static void vulcan_shot_action(SHIP_MGR *mgr, SHOT_MGR *shot_mgr, SHIP_MGR *enemy_mgr) {

	if (input[P1] & INPUT_REPEAT_BUTTON1) {
		SHOT_INFO *shot = shot_new(shot_mgr, SHOT_ID_NORMAL);
		if (shot != NULL) {
			shot_fire(shot, ship, 0, -32);
			shot->sprite.dy = IntToReal(-16);
		}
	}
}

/* eCVbg */
static void tail_shot_action(SHIP_MGR *mgr, SHOT_MGR *shot_mgr, SHIP_MGR *enemy_mgr) {

	if (shot_mgr->active == NULL && input[P1] & INPUT_REPEAT_BUTTON2) {
		double angle = 0;
		Real speed = IntToReal(8);
		SHOT_INFO *shot1 = shot_new(shot_mgr, SHOT_ID_MBALLET);
		SHOT_INFO *shot2 = shot_new(shot_mgr, SHOT_ID_MBALLET);
		SHOT_INFO *shot3 = shot_new(shot_mgr, SHOT_ID_MBALLET);
		SHOT_INFO *shot4 = shot_new(shot_mgr, SHOT_ID_MBALLET);
		SHOT_INFO *shot5 = shot_new(shot_mgr, SHOT_ID_MBALLET);

		if (shot1 != NULL) {
			shot_fire(shot1, ship, 0, 0);
			sprite_set_angle(&shot1->sprite, angle, speed);
		}
		if (shot2 != NULL) {
			shot_fire(shot2, ship, 0, 0);
			sprite_set_angle(&shot2->sprite, angle - 15, speed);
		}
		if (shot3 != NULL) {
			shot_fire(shot3, ship, 0, 0);
			sprite_set_angle(&shot3->sprite, angle + 15, speed);
		}
		if (shot4 != NULL) {
			shot_fire(shot4, ship, 0, 0);
			sprite_set_angle(&shot4->sprite, angle - 30, speed);
		}
		if (shot5 != NULL) {
			shot_fire(shot5, ship, 0, 0);
			sprite_set_angle(&shot5->sprite, angle + 30, speed);
		}
	}
}

/* t@C[\[h */
static void fire_sword_action(SHIP_MGR *mgr, SHOT_MGR *shot_mgr, SHIP_MGR *enemy_mgr) {

	/* t@C[\[h̓͐ */
	if (!(ship->flag & FIRE_SWORD)) {
		if (!(ship->flag & SEARCH_LASER)) {
			if (input[P1] & INPUT_PRESS_BUTTON1) {

				if (sword_enabled < SWORD_WAIT) {
					++sword_enabled;
				} else {
					if (sword < SWORD_POWER_MAX) ++sword;
					input[P1] &= ~INPUT_REPEAT_BUTTON1;
				}
			}

			if (input[P1] & INPUT_KEYUP_BUTTON1) {
				ship->flag |= FIRE_SWORD;
				sword *= 2;
				sword_enabled = 0;
			}
		}
	}

	/* t@C[\[h̐ */
	if (ship->flag & FIRE_SWORD) {
		if (sword_enabled < HEIGHT) sword_enabled += 64;

		if (--sword < 0) {
			ship->flag &= ~FIRE_SWORD;
			sword_enabled = 0;
			sword = 0;

		} else {
			SHIP_INFO *enemy;

			int x = RealToInt(ship->sprite.x) + 30;
			int y = RealToInt(ship->sprite.y) + 24;

			int s = (sword > 100)? 100: sword;
			int w = FIRE_SWORD_W * s / 200;
			int x1 = x - w;
			int x2 = x + w;
			int y1 = y - sword_enabled;

			/* G@Ƃ̐ڐG */
			for (enemy = enemy_mgr->active; enemy != NULL; enemy = enemy->next) {

				if (enemy->stat != SHIP_STAT_LOST
					&& enemy->stat != SHIP_STAT_EXPLOSION
					&& sprite_rect_check(&enemy->sprite, x1, y1, x2, y) == SDL_TRUE) {

					int strength = STRENGTH_SWORD;
					system_info.score += ship_down(enemy, &strength, 0, 0);
					if (enemy->strength > 0) {
						effect_set_explosion(&enemy->sprite);
					}
				}
			}
#if 1 /* GeȂꍇ́A0ɕύX */
			/* GeƂ̐ڐG */
			if (sword >= 100) {
				SHOT_INFO *prev = NULL;
				SHOT_INFO *info = enemy_shot_mgr.active;

				while (info != NULL) {
					if (sprite_rect_check(&info->sprite, x1, y1, x2, y) == SDL_TRUE) {
						SHOT_INFO *next = info->next;

						effect_set_explosion(&info->sprite);
						info->strength -= STRENGTH_SWORD;
						if (info->strength <= 0) {
							system_info.score++;
							shot_delete(&enemy_shot_mgr, info, prev);
							info = next;
							continue;
						}
					}
					prev = info;
					info = info->next;
				}
			}
#endif
		}

	} else if (!(input[P1] & INPUT_PRESS_BUTTON2)) {
		/* oJVbg */
		vulcan_shot_action(mgr, shot_mgr, enemy_mgr);
	}
}

static void myship_search_enemy(void) {

	int x = RealToInt(ship->sprite.x) + 30;
	int y = RealToInt(ship->sprite.y) + 24;
	SHIP_INFO *enemy;
	int top = 0;
	int r = radar * radar;

	/* G@Ƃ̐ڐG */
	for (enemy = enemy_mgr.active; enemy != NULL; enemy = enemy->next) {
		SHIP_TYPE *p = &ship_type[enemy->id];
		SPRITE *s = &enemy->sprite;

		if (p->lockon >= 0
			&& enemy->stat != SHIP_STAT_LOST
			&& enemy->stat != SHIP_STAT_EXPLOSION
			&& !(enemy->flag & SHIP_FLAG_NODAMAGE)
			&& s->options & SPRITE_USE_COLLISION) {

			if (!(enemy->flag & SHIP_FLAG_LOCKON) || (enemy->flag & SHIP_FLAG_HEAVY)) {

				int x1 = RealToInt(s->x + ((s->collision.x1 - s->collision.x0) >> 1)) - x;
				int y1 = RealToInt(s->y + ((s->collision.y1 - s->collision.y0) >> 1)) - y;
				int x2 = x1 * x1;
				int y2 = y1 * y1;
				int r2 = x2 + y2;

				if (r2 <= r) {
					int i = top++;

					for (; i > 0; i--) {
						if (range_infos[i-1].range <= r2) break;

						range_infos[i].ship = range_infos[i-1].ship;
						range_infos[i].range = range_infos[i-1].range;
					}

					range_infos[i].ship = enemy;
					range_infos[i].range = r2;
				}
			}
		}
	}

	if (top > 0) {
		range_infos[0].ship->flag |= SHIP_FLAG_LOCKON;
		lockon[lockon_cnt++] = range_infos[0];
	}
}

static void myship_lockon_action(void) {

	int i;

	for (i = 0; i < lockon_cnt;) {
		SHIP_INFO *enemy = lockon[i].ship;

		if (enemy->stat != SHIP_STAT_LOST
			&& enemy->stat != SHIP_STAT_EXPLOSION) {
			++i;

		} else {
			int j;
			for (j = i + 1; j < lockon_cnt; j++) {
				lockon[j - 1] = lockon[j];
			}
			--lockon_cnt;
		}
	}
}

/* T[`[U[ */
static void search_laser_action(SHIP_MGR *mgr, SHOT_MGR *shot_mgr, SHIP_MGR *enemy_mgr) {

	/* UAbNIO */
	int i;
	for (i = 0; i < lockon_cnt; i++) {
		SHIP_INFO *enemy = lockon[i].ship;
		enemy->flag &= ~SHIP_FLAG_LOCKON;
	}
	lockon_cnt = 0;

	/* T[`[U[̓͐iTYPE-1j */
	if (input[P1] & INPUT_PRESS_BUTTON2) {

		if (radar < RADAR_R) {
			/* [_[WJ */
			radar += RADAR_SPEED;
		}
		if (lockon_cnt < LOCKON_MAX) {
			myship_search_enemy();
		}
		ship->flag |= SEARCH_LASER;

	} else {
		ship->flag &= ~SEARCH_LASER;
		radar = 0;
		lockon_cnt = 0;
	}

	if (ship->flag & SEARCH_LASER) {
		if (lockon_cnt > 0) {
			int i = radar_enabled++;
			int target = i % lockon_cnt;
			SHIP_INFO *enemy = lockon[target].ship;

			if (enemy != NULL) {
				double strength = lockon[target].range / (double)(RADAR_R * RADAR_R);
				int power = (int)(STRENGTH_SEARCH * (1 - strength)) + 1;
				//font_locate(0, 1);
				//font_printf("range=%8d\n", lockon[target].range);
				//font_printf("strength=%1.4f\n", strength);
				//font_printf("power=%2d\n", power);

				system_info.score += ship_down(enemy, &power, 0, 0);
				if (enemy->strength > 0) {
					effect_set_explosion(&enemy->sprite);
				}
			}
		}

	} else {
		/* oJVbg */
		vulcan_shot_action(mgr, shot_mgr, enemy_mgr);
	}
}

/* bNI[U[ */
static void lockon_laser_action(SHIP_MGR *mgr, SHOT_MGR *shot_mgr, SHIP_MGR *enemy_mgr) {

	/* [U[̓͐iTYPE-2j */
	if (!(ship->flag & SEARCH_LASER)) {
		if (input[P1] & INPUT_PRESS_BUTTON2) {

			if (radar < RADAR_R * 5 / 4) {
				/* [_[WJ */
				radar += RADAR_SPEED * 5 / 4;
			}
			if (lockon_cnt < LOCKON_MAX) {
				myship_search_enemy();
			}
		}

		if (input[P1] & INPUT_KEYUP_BUTTON2) {
			radar = 0;
			if (lockon_cnt > 0) {
				ship->flag |= SEARCH_LASER;
				radar_enabled = LASER_TIME_MAX;
			}
		}
	}

	if (ship->flag & SEARCH_LASER) {
		if (radar_enabled-- < 0 || lockon_cnt <= 0) {
			int i;
			for (i = 0; i < lockon_cnt; i++) {
				SHIP_INFO *enemy = lockon[i].ship;
				enemy->flag &= ~SHIP_FLAG_LOCKON;
			}
			ship->flag &= ~SEARCH_LASER;
			radar = 0;
			lockon_cnt = 0;
			radar_enabled = 0;

		} else if (frame_count & 1) {
			int i = radar_enabled++;
			SHIP_INFO *enemy = lockon[i % lockon_cnt].ship;

			if (enemy != NULL) {
				int strength = STRENGTH_LOCKON;
				system_info.score += ship_down(enemy, &strength, 0, 0);

				if (enemy->strength > 0) {
					effect_set_explosion(&enemy->sprite);
				}
			}
		}

	} else {
		/* oJVbg */
		vulcan_shot_action(mgr, shot_mgr, enemy_mgr);
	}
}

/* z[~O[U[ */
static void homing_laser_action(SHIP_MGR *mgr, SHOT_MGR *shot_mgr, SHIP_MGR *enemy_mgr) {

	/* [U[̓͐iTYPE-3j */
	if (input[P1] & INPUT_PRESS_BUTTON2) {

		if (radar < RADAR_R) {
			/* [_[WJ */
			radar += RADAR_SPEED * 2;
		}
		if (lockon_cnt < LOCKON_MAX) {
			myship_search_enemy();
		}
		ship->flag |= SEARCH_LASER;
	}

	if (input[P1] & INPUT_KEYUP_BUTTON2) {
		int i;

		if (lockon_cnt > 0) {
			for (i = 0; i < lockon_cnt; i++) {
				SHIP_INFO *enemy = lockon[i].ship;
				SHOT_INFO *shot = shot_new(shot_mgr, SHOT_ID_PHOTON_S);

				if (shot != NULL) {
					int s = (i >> 1) + 1;
					double angle = ((i & 1)? s: -s) * 9;
					shot_fire(shot, ship, 0, 30);
					sprite_set_angle(&shot->sprite, angle, IntToReal(8 + s * 2));
					shot->target = enemy;
				}
			}

		} else {
			for (i = 0; i < LOCKON_MAX / 2; i++) {
				SHOT_INFO *shot = shot_new(shot_mgr, SHOT_ID_PHOTON_D);

				if (shot != NULL) {
					int s = (i >> 1) + 4;
					double angle = ((i & 1)? s: -s) * 20;
					shot_fire(shot, ship, 0, 30);
					sprite_set_angle(&shot->sprite, angle, IntToReal(5));
				}
			}

		}

		radar = 0;
		lockon_cnt = 0;
		ship->flag &= ~SEARCH_LASER;
	}

	if (!(ship->flag & SEARCH_LASER)) {
		/* oJVbg */
		vulcan_shot_action(mgr, shot_mgr, enemy_mgr);
	}
}

/* @̏ */
void myship_init(SHIP_MGR *mgr) {
	ship = ship_new(mgr, SHIP_ID_ME);
	ship->sprite.x = IntToReal(WIDTH / 2 - 30);
	ship->sprite.y = IntToReal(HEIGHT    - 60);

	sword = 0;
	sword_enabled = 0;

	radar = 0;
	radar_enabled = 0;
	lockon_cnt = 0;
	/*laser_type = LASER_TYPE_HOMING; /**/

	memset(range_infos, 0, sizeof(range_infos));
	memset(lockon, 0, sizeof(lockon));
}

/* @̐ */
void myship_action(SHIP_MGR *mgr, SHOT_MGR *shot_mgr, SHIP_MGR *enemy_mgr) {

	/* @̑ */
	Real spd = (laser_type == LASER_TYPE_SWORD
				|| !(input[P1] & INPUT_PRESS_BUTTON2)) ?
		IntToReal(MYSHIP_SPEED): IntToReal(MYSHIP_SPEED/2);

	if (input[P1] & INPUT_PRESS_LEFT ) ship->sprite.x -= spd;
	if (input[P1] & INPUT_PRESS_RIGHT) ship->sprite.x += spd;
	if (input[P1] & INPUT_PRESS_UP   ) ship->sprite.y -= spd;
	if (input[P1] & INPUT_PRESS_DOWN ) ship->sprite.y += spd;

	if (ship->sprite.x < IntToReal(-SIDE_SPACE))           ship->sprite.x = IntToReal(-SIDE_SPACE);
	if (ship->sprite.x > IntToReal(WIDTH - SIDE_SPACE))    ship->sprite.x = IntToReal(WIDTH - 30);
	if (ship->sprite.y < 0)                                ship->sprite.y = 0;
	if (ship->sprite.y > IntToReal(HEIGHT - BOTTOM_SPACE)) ship->sprite.y = IntToReal(HEIGHT - BOTTOM_SPACE);

	if (ship->stat != SHIP_STAT_LOST) {

		/* bNI */
		myship_lockon_action();

		switch (laser_type) {
		case LASER_TYPE_SWORD:
			/* t@C[\[h̐ */
			fire_sword_action(mgr, shot_mgr, enemy_mgr);

			/* eCVbg */
			tail_shot_action(mgr, shot_mgr, enemy_mgr);
			break;

		case LASER_TYPE_SEARCH:
			/* T[`[U[̐ */
			search_laser_action(mgr, shot_mgr, enemy_mgr);
			break;

		case LASER_TYPE_LOCKON:
			/* bNI[U[̐ */
			lockon_laser_action(mgr, shot_mgr, enemy_mgr);
			break;

		default:
		case LASER_TYPE_HOMING:
			/* T[`[U[̐ */
			homing_laser_action(mgr, shot_mgr, enemy_mgr);
			break;
		}
	}

	/* XvCg̐ */
	sprite_action(&ship->sprite);

	/* Ԃɉo */
	switch (ship->stat) {
	case SHIP_STAT_EXPLOSION:
		if (ship->count == 0) {
			sound_stop_effect();
		}
		sound_effect(SE_BOMB2);
		if (ship->count++ > 2) {
			ship->stat = SHIP_STAT_LOST;
			system_info.status &= ~SYSTEM_MODE_MASK;
			system_info.status |= SYSTEM_MODE_OVER;
			sound_music(BGM_03, -1);
		}
		break;

	case SHIP_STAT_LOST:
		break;
	}

	/*ship->stat = SHIP_STAT_NORMAL; /**/
}

void myship_rendaring(void) {

	if (ship->stat != SHIP_STAT_LOST) {
		/* SWORD̕\ */
		if (ship->flag & FIRE_SWORD) {

			int x = RealToInt(ship->sprite.x);
			int y = RealToInt(ship->sprite.y) + 24;
			int s = (sword > 100)? 100: sword;
			int w = FIRE_SWORD_W * s / 200;
			int x0 = x + 30;


			draw_trect(x0 - w, y - sword_enabled, x0 - w + 2, y, 0xf8ffffff);
			draw_srect(x0 - w + 3, y - sword_enabled, x0 + w - 3, y, 0xf8ff0000);
			draw_trect(x0 + w - 2, y - sword_enabled, x0 + w, y, 0xf8ffffff);
		}

		/* XvCg̕\ */
		/* @\B */
		sprite_draw(screen, &ship->sprite);

		if (laser_type == LASER_TYPE_SEARCH
			|| laser_type == LASER_TYPE_LOCKON) {

			/* [U[̕\ */
			if ((ship->flag & SEARCH_LASER) && lockon_cnt > 0) {
				int x1 = RealToInt(ship->sprite.x) + 20;
				int x2 = RealToInt(ship->sprite.x) + 40;
				int y = RealToInt(ship->sprite.y) + 30;
				int i;

				for (i = 0; i < LOCKON_MAX; i++) {
					SHIP_INFO *enemy = lockon[i % lockon_cnt].ship;
					SPRITE *s = &enemy->sprite;

					int x0 = RealToInt(s->x + s->collision.x0
									   + ((s->collision.x1 - s->collision.x0) >> 1));
					int y0 = RealToInt(s->y + s->collision.y0
									   + ((s->collision.y1 - s->collision.y0) >> 1));
					int b = i % (LOCKON_MAX/2) * 4;
					int x = (i & 1)? x2 + b: x1 - b;

					draw_line(x, y, x0, y0, 3, 0x00ffe000);
				}
			}
		}

		/* [_[̕\ */
		if (radar > 0) {
			draw_circle(RealToInt(ship->sprite.x) + 30,
						RealToInt(ship->sprite.y) + 32,
						radar, 3, 0x0000ff00);
		}
	}

	/* SWORDQ[W̕\ */
	if (sword > 0) {
		int x = WIDTH * 10 / 100;
		int w = WIDTH - x - FONT_MAP_SIZE * 11;

		if (ship->flag & FIRE_SWORD) {
			w = w * sword / 200;
			draw_rect(x, HEIGHT - 16, x + w, HEIGHT - 14, 0x00ff0000);

		} else {
			w = w * sword / 100;
			draw_rect(x, HEIGHT - 16, x + w, HEIGHT - 14, 0x0000ff00);
		}
	}
}
