/*
 * rankin.c
 *
 * Last modified: <2009/11/28 21:41:04 +0900> By SUMIDA
 */
#include <SDL.h>
#include <search.h>
#include "system.h"
#include "bg.h"
#include "texture.h"
#include "input.h"

#define RANKING_MAX  10
#define CHARA_WIDTH  15
#define CHARA_HEIGHT 10

extern char datpath[_MAX_PATH];
static int comp_info(const void *d0, const void *d1);
static int find_rank(void);

PLAYER_INFO ranking[RANKING_MAX];
PLAYER_INFO default_ranking[] = {
	{10000, LASER_TYPE_SWORD , 0, 5, "MIRROR    "},
	{ 9000, LASER_TYPE_SEARCH, 0, 5, "ALMA!     "},
	{ 8000, LASER_TYPE_LOCKON, 0, 4, "SAMANSA   "},
	{ 7000, LASER_TYPE_HOMING, 0, 4, "TABASA    "},
	{ 6000, LASER_TYPE_SWORD , 0, 3, "Go!Go!Go! "},
	{ 5000, LASER_TYPE_SEARCH, 0, 3, "HooBar    "},
	{ 4000, LASER_TYPE_LOCKON, 0, 2, "LuckAnd..."},
	{ 3000, LASER_TYPE_HOMING, 0, 2, "Guest     "},
	{ 2000, LASER_TYPE_SWORD , 0, 1, "HogeHoge  "},
	{ 1000, LASER_TYPE_SEARCH, 0, 1, "LookAtMe! "},
};

int rank = -1;
static int column;
static int chara;

const static chara_tbl[] = {
	0x41,0x42,0x43,0x44,0x45,0x46,0x47,0x48,0x49,0x4A,0x4B,0x4C,0x4D,0x4E,0x4F,
	0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5A,0x61,0x62,0x63,0x64,
	0x65,0x66,0x67,0x68,0x69,0x6A,0x6B,0x6C,0x6D,0x6E,0x6F,0x70,0x71,0x72,0x73,
	0x74,0x75,0x76,0x77,0x78,0x79,0x7A,0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,
	0x38,0x39,0x7B,0x7C,0x7D,0x7E,0xAE,0xAF,0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,
	0xE7,0xE8,0xE9,0xEA,0xEB,0xEC,0xED,0xEE,0xEF,0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,
	0xF6,0xF7,0xF8,0xF9,0x01,0x02,0x03,0x04,0x05,0x06,0x0B,0x0C,0x0E,0x0F,0x10,
	0x11,0x12,0x13,0x14,0x15,0x17,0x18,0x19,0x1A,0x1B,0x1D,0x1E,0x1F,0x21,0x22,
	0x23,0x24,0x25,0x26,0x27,0x28,0x60,0x29,0x2A,0x2B,0x2C,0x2D,0x2E,0x2F,0x3A,
	0x3B,0x3C,0x3D,0x3E,0x3F,0x40,0x5B,0x5C,0x5D,0x5E,0x5F,0x20,
};


int rankin_init(void) {
	player_info.score = system_info.score;
	player_info.finish_level = system_info.level;
	snprintf(player_info.name, sizeof(player_info.name), "%10c", ' ');

	column = 0;
	chara = 0;

	qsort(ranking, NUMBEROF(ranking), sizeof(PLAYER_INFO), comp_info);
	rank = find_rank();

	if (rank >= 0) {
		int i;
		for (i = NUMBEROF(ranking) - 1; i > rank; i--) {
			ranking[i] = ranking[i - 1];
		}
		ranking[rank] = player_info; /* ݒ */
	}
	return rank;
}

int rankin_action(void) {

	++system_info.count;

	if (column < NAME_LENGTH) {
		int curx = chara % CHARA_WIDTH;
		int cury = chara / CHARA_WIDTH;
		if (input[P1] & INPUT_KEYDOWN_RIGHT) ++curx;
		if (input[P1] & INPUT_KEYDOWN_LEFT ) --curx;
		if (input[P1] & INPUT_KEYDOWN_UP   ) --cury;
		if (input[P1] & INPUT_KEYDOWN_DOWN ) ++cury;
		if (curx <  0) curx = CHARA_WIDTH - 1;
		if (curx > CHARA_WIDTH  - 1) curx =  0;
		if (cury <  0) cury = CHARA_HEIGHT -1;
		if (cury > CHARA_HEIGHT - 1) cury =  0;
		chara = curx + cury * CHARA_WIDTH;
		if (chara >= NUMBEROF(chara_tbl)) chara = NUMBEROF(chara_tbl);
	} else {
		chara = NUMBEROF(chara_tbl);
	}

	if (input[P1] & INPUT_KEYDOWN_BUTTON1) {
		if (chara >= NUMBEROF(chara_tbl)) {
			ranking[rank] = player_info;
			return 1;
		}

		player_info.name[column++] = chara_tbl[chara];
	}

	if (input[P1] & INPUT_KEYDOWN_BUTTON2 && column > 0) {
		player_info.name[--column] = ' ';
	}

	return 0;
}

void rankin_draw(void) {
	int i, j;

	font_clean();
	font_default_frame();

	font_locate(2, 3);
	font_texture(TEX_FONT_RED);
	font_printf("Name entry");

	font_texture(TEX_FONT_GREEN);
	font_locate(5, 5);
	font_printf("You are No.%d player!", rank + 1);

	font_frame(5, 8, 31, 22);
	font_texture(TEX_FONT_BLUE);
	font_locate(0, 0);
	font_printf("                               "
				" \n \n \n \n \n \n \n \n \n \n"
				"                               ");

	font_frame(6, 9, 30, 20);
	font_locate(0, 0);
	font_attribute(0);
	for (i = 0; i < NUMBEROF(chara_tbl); i++) {
		font_printf("%c ", chara_tbl[i]);
	}
	font_locate(24, 9);
	font_printf("  End ");

	font_locate(0, 12);
	font_printf("Please enter your name:");

	font_default_texture();
	if (column < NAME_LENGTH) {
		int c = 8 + column;
		font_locate(c, 14);
		font_printf("%c", 0x1F);
		font_locate(c, 16);
		font_printf("%c", 0x1E);
	}

	if (system_info.count & 1)
		font_texture(TEX_FONT_YELLOW);
	else
		font_default_texture();

	font_locate(8, 15);
	font_printf("%s", player_info.name);

	if (chara == NUMBEROF(chara_tbl)) {
		font_locate(25, 9);
		font_printf("[");
		font_locate(29, 9);
		font_printf("]");
	} else {
		int curx = (chara % CHARA_WIDTH) * 2;
		int cury = chara / CHARA_WIDTH;
		font_locate(curx - 1, cury);
		font_printf("[");
		font_locate(curx + 1, cury);
		font_printf("]");
	}

	font_default_frame();
	font_attribute(1);
}

int comp_info(const void *d0, const void *d1) {
	const PLAYER_INFO *p0 = d0;
	const PLAYER_INFO *p1 = d1;
	return p1->score - p0->score;
}

int find_rank(void) {
	int result = -1;
	int left  = 0;
	int right = NUMBEROF(ranking) - 1;
	Uint32 score = player_info.score;

	while (left <= right) {
		int center = (left + right) / 2;
		PLAYER_INFO *info = &ranking[center];

		if (score > info->score) {
			right = center - 1;
			if (right < 0) {
				result = 0;
				break;
			}

		} else if (score < info->score) {
			left = center + 1;
			if (left >= NUMBEROF(ranking)) break;

		} else {
			for (; center < NUMBEROF(ranking); center++) {
				if (score > ranking[center].score) {
					result = center;
					break;
				}
			}
			break;
		}

		if (left > right) {
			result = left;
			break;
		}
	}

	/*result = 2; /* debug */
	return result;
}

/* XRȀ */
void score_init(void) {
	memcpy(&ranking, &default_ranking, sizeof(ranking));
}

/* XRA̓ǂݍ */
void score_load(void) {

	FILE *fp = fopen(datpath, "r");

	if (fp != NULL) {
		char line[256];
		int i = 0;

		while (fgets(line, sizeof(line), fp) != NULL) {

			if (line[0] != '#') {
				PLAYER_INFO *info = &ranking[i];
				char sp[] = {255, 0};
				char *val;

				/* XRA */
				if ((val = strtok(line, sp)) != NULL) {
					info->score = atoi(val);
				}

				/* XyVEF| */
				if ((val = strtok(NULL, sp)) != NULL) {
					info->laser_type = atoi(val) % LASER_TYPE_MAX;
				}

				/* Q[Jñx */
				if ((val = strtok(NULL, sp)) != NULL) {
					info->start_level = (char)atoi(val);
				}

				/* Q[Ĩx */
				if ((val = strtok(NULL, sp)) != NULL) {
					info->finish_level = (char)atoi(val);
				}

				/* vC[ */
				if ((val = strtok(NULL, sp)) != NULL) {
					snprintf(info->name, sizeof(info->name), "%10.10s", val);
				}

				REPORT(format("%d: score=%d, type=%d, level=%d-%d, name=%s",
							  i,
							  info->score,
							  info->laser_type,
							  info->start_level,
							  info->finish_level,
							  info->name));

				if (++i >= NUMBEROF(ranking)) break;
			}
		}

		fclose(fp);

	} else {
		perror(datpath);
	}
}

/* XRA̕ۑ */
void score_save(void) {

	FILE *fp = fopen(datpath, "w");

	if (fp != NULL) {
		int i;

		for (i = 0; i < NUMBEROF(ranking); i++) {
			PLAYER_INFO *info = &ranking[i];
			char sp = 255;

			fprintf(fp,
					"%d%c"
					"%d%c"
					"%d%c"
					"%d%c"
					"%s\n",
					info->score, sp,
					info->laser_type, sp,
					info->start_level, sp,
					info->finish_level, sp,
					info->name);
		}

		fclose(fp);

	} else {
		perror(datpath);
	}
}

/* XRALO̕\ */
void score_draw(void) {
	const static char *spw_shortname[] = {
		"FSWORD", "SEARCH", "LOCKON", "HOMING"};
	static char line[34];
	static int len = sizeof(line) - 1;
	int i;

	font_clean();

	font_locate(2, 2);
	font_texture(TEX_FONT_RED);
	font_printf("Score ranking");

	font_locate(2, 4);
	font_texture(TEX_FONT_BLUE);
	font_printf("No Name       Score     Type   Level");

	font_default_texture();
	font_attribute(0);

	for (i = 0; i < NUMBEROF(ranking); i++) {
		PLAYER_INFO *p = &ranking[i];
		int l = system_info.count / len;

		if (rank == i && system_info.count & 1)
			font_texture(TEX_FONT_YELLOW);
		else
			font_default_texture();

		if (i < l) {
			font_locate(2, i * 2 + 6);
			font_printf("%2d %10s %9d %6s %02d-%02d",
						i + 1, p->name, p->score, spw_shortname[p->laser_type],
						p->start_level, p->finish_level);
		} else if (i == l) {
			int c = system_info.count % len;
			font_locate(2 + len - c, i * 2 + 6);
			snprintf(line, sizeof(line), "%2d %10s %9d %6s %02d-%02d",
					 i + 1, p->name, p->score, spw_shortname[p->laser_type],
					 p->start_level, p->finish_level);
			font_printf("%.*s", c, line);
		}
	}

	font_attribute(1);
}
