/*
 * ship.c
 *
 * Last modified: <2009/11/26 12:21:44 +0900> By SUMIDA
 */
#include <stdlib.h>
#include <string.h>
#include <SDL.h>
#include "system.h"
#include "texture.h"
#include "surface.h"
#include "sprite.h"
#include "ship.h"
#include "effect.h"
#include "sound.h"
#include "shot.h"
#include "enemy.h"

#define SPRITE_OPT_SHIP  SPRITE_VISIBLE|SPRITE_FRAME|SPRITE_LOOP_ANIME|SPRITE_USE_ANIME|SPRITE_USE_MOVE|SPRITE_USE_COLLISION
#define SPRITE_OPT_SIGHT SPRITE_VISIBLE|SPRITE_FRAME64
#define SPRITE_OPT_FORCE SPRITE_VISIBLE|SPRITE_FRAME64|SPRITE_LOOP_ANIME|SPRITE_USE_ANIME|SPRITE_USE_COLLISION
#define SPRITE_OPT_ROCK  SPRITE_VISIBLE|SPRITE_USE_MOVE|SPRITE_USE_COLLISION

SHIP_TYPE ship_type[] = {
	{TEX_SHIP  ,     1,    0, -1, SPRITE_OPT_SHIP , {IntToReal(28), IntToReal(20), IntToReal(60-29) , IntToReal(60-21)}},
	{TEX_ENEMY1,     1,   10,  1, SPRITE_OPT_SHIP , {IntToReal(2) , IntToReal(2) , IntToReal(60-3)  , IntToReal(60-3) }},
	{TEX_ENEMY2,     2,   20,  2, SPRITE_OPT_SHIP , {IntToReal(2) , IntToReal(2) , IntToReal(60-3)  , IntToReal(60-3) }},
	{TEX_ENEMY3,    10,   30,  3, SPRITE_OPT_SHIP , {IntToReal(2) , IntToReal(2) , IntToReal(60-3)  , IntToReal(60-3) }},
	{TEX_ENEMY4,    30,   50,  4, SPRITE_OPT_SHIP , {IntToReal(2) , IntToReal(2) , IntToReal(60-3)  , IntToReal(60-3) }},
	{TEX_ENEMY5,    20,   40,  5, SPRITE_OPT_SHIP , {IntToReal(2) , IntToReal(2) , IntToReal(60-3)  , IntToReal(60-3) }},
	{TEX_MISSILE,    1,   10,  0, SPRITE_OPT_SHIP , {IntToReal(4) , IntToReal(10), IntToReal(16-4)  , IntToReal(54-1) }},
	{TEX_ROCK  ,    50,  100, -1, SPRITE_OPT_ROCK , {IntToReal(4) , IntToReal(10), IntToReal(80-4)  , IntToReal(64-4) }},
	{TEX_ROCK_M1,   25,   25, -1, SPRITE_OPT_ROCK , {IntToReal(2) , IntToReal(2) , IntToReal(32-2)  , IntToReal(32-2) }},
	{TEX_ROCK_M2,   25,   25, -1, SPRITE_OPT_ROCK , {IntToReal(2) , IntToReal(2) , IntToReal(32-2)  , IntToReal(32-2) }},
	{TEX_ROCK_M3,   25,   25, -1, SPRITE_OPT_ROCK , {IntToReal(2) , IntToReal(2) , IntToReal(32-2)  , IntToReal(32-2) }},
	{TEX_ROCK_S1,   10,    5, -1, SPRITE_OPT_ROCK , {IntToReal(1) , IntToReal(1) , IntToReal(16-1)  , IntToReal(16-1) }},
	{TEX_ROCK_S2,   10,    5, -1, SPRITE_OPT_ROCK , {IntToReal(1) , IntToReal(1) , IntToReal(16-1)  , IntToReal(16-1) }},
	{TEX_ROCK_S3,   70,    5, -1, SPRITE_OPT_ROCK , {IntToReal(1) , IntToReal(1) , IntToReal(16-1)  , IntToReal(16-1) }},
	{TEX_BOSS1 ,   500, 1000,  6, SPRITE_OPT_SHIP , {IntToReal(18), IntToReal(45), IntToReal(100-19), IntToReal(164)  }},
	{TEX_BOSS2 ,  1000, 5000,  7, SPRITE_OPT_SHIP , {IntToReal(8) , IntToReal(42), IntToReal(100-9) , IntToReal(116)  }},
};

extern SHIP_MGR enemy_mgr;

void ship_init(SHIP_MGR *mgr, SHIP_INFO *infos, size_t number) {

	size_t i;

	if (mgr != NULL) {
		memset(mgr, 0, sizeof(SHIP_MGR));
		mgr->pool = infos;
	}

	memset(infos, 0, sizeof(SHIP_INFO) * number);

	for (i = 0; i < number - 1; i++) {
		infos[i].next = &infos[i + 1];
	}
}


SHIP_INFO *ship_new(SHIP_MGR *mgr, SHIP_ID id) {

	SHIP_INFO *info = mgr->pool;

	if (info != NULL) {
		SHIP_TYPE *p = &ship_type[id];

		mgr->pool   = info->next;
		info->next  = mgr->active;
		mgr->active = info;

		info->id       = id;
		info->stat     = SHIP_STAT_NORMAL;
		info->flag     = 0;
		info->count    = 0;
		info->point    = p->point;
		info->strength = p->strength;
		info->sprite.surface_id = p->tex_id;
		info->sprite.options = p->options;
		info->sprite.frame_start    = 0;
		info->sprite.frame_end      = 1;
		info->sprite.frame_interval = 5;
		info->sprite.collision.x0   = p->collision.x0;
		info->sprite.collision.y0   = p->collision.y0;
		info->sprite.collision.x1   = p->collision.x1;
		info->sprite.collision.y1   = p->collision.y1;
	}

	return info;
}

void ship_delete(SHIP_MGR *mgr, SHIP_INFO *info, SHIP_INFO *prev) {

//	if (!(info->flag & SHIP_FLAG_NOPOOL)) {

		if (prev == NULL) {
			mgr->active = info->next;
		} else {
			prev->next = info->next;
		}

		memset(info, 0, sizeof(SHIP_INFO));
		info->next = mgr->pool;
		mgr->pool = info;
//	}
}

int ship_down(SHIP_INFO *info, int *strength, Real dx, Real dy) {

	int score = 0;

	if (info->stat != SHIP_STAT_EXPLOSION && info->stat != SHIP_STAT_LOST) {

		int pow = (*strength > SHOT_LIMIT_STRENGTH)? SHOT_LIMIT_STRENGTH: *strength;

		*strength -= pow;
		if (!(info->flag & SHIP_FLAG_NODAMAGE)) {
			info->strength -= pow;
		}

		if (info->strength > 0) {
			info->flag |= SHIP_FLAG_DOWN;
			if (!(info->flag & SHIP_FLAG_HEAVY)) {
				info->sprite.x += dx;
				info->sprite.y += dy;
			}
			sound_effect(SE_SHOT1);

		} else {
			EFFECT_INFO *effect;

			REPORT(format("bomb! (info=%p, id=%d, point=%d)", info, info->id, info->point));
			score = info->point;

			info->flag &= ~SHIP_FLAG_LOCKON;
			info->stat = SHIP_STAT_EXPLOSION;
			info->count = SHIP_DEF_LOST_WAIT;
			info->sprite.options &= ~SPRITE_USE_MOVE;

			switch (info->id) {
			case SHIP_ID_BOSS1:
			case SHIP_ID_BOSS2:
				effect = effect_new(EFFECT_ID_BOMB);
				sound_effect(SE_BOMB1);
				break;

			case SHIP_ID_ROCK:
				{
					SHIP_INFO *r1 = enemy_new(&enemy_mgr, SHIP_ID_ROCK_M1,
											  info->sprite.x + ToReal(Random(0, 16)),
											  info->sprite.y);

					SHIP_INFO *r2 = enemy_new(&enemy_mgr, SHIP_ID_ROCK_M2,
											  info->sprite.x + ToReal(Random(16, 32) + 16),
											  info->sprite.y);

					SHIP_INFO *r3 = enemy_new(&enemy_mgr, SHIP_ID_ROCK_M3,
											  info->sprite.x + ToReal(Random(32, 48) + 32),
											  info->sprite.y);

					if (r1 != NULL) {
						r1->sprite.dx = (r1->sprite.dx + info->sprite.dx) >> 1;
						r1->sprite.dy = (r1->sprite.dy + info->sprite.dy) >> 1;
					}
					if (r2 != NULL) {
						r2->sprite.dx = (r2->sprite.dx + info->sprite.dx) >> 1;
						r2->sprite.dy = (r2->sprite.dy + info->sprite.dy) >> 1;
					}
					if (r3 != NULL) {
						r3->sprite.dx = (r3->sprite.dx + info->sprite.dx) >> 1;
						r3->sprite.dy = (r3->sprite.dy + info->sprite.dy) >> 1;
					}
					effect = effect_new(EFFECT_ID_EXPLOSION);
					sound_effect(SE_BOMB0);
				}
				break;

			case SHIP_ID_ROCK_M1:
			case SHIP_ID_ROCK_M2:
			case SHIP_ID_ROCK_M3:
				{
					SHIP_INFO *r1 = enemy_new(&enemy_mgr, SHIP_ID_ROCK_S1,
											  info->sprite.x + ToReal(Random(0, 4)),
											  info->sprite.y);

					SHIP_INFO *r2 = enemy_new(&enemy_mgr, SHIP_ID_ROCK_S2,
											  info->sprite.x + ToReal(Random(4, 8) + 4),
											  info->sprite.y);

					SHIP_INFO *r3 = enemy_new(&enemy_mgr, SHIP_ID_ROCK_S3,
											  info->sprite.x + ToReal(Random(8, 16) + 8),
											  info->sprite.y);

					if (r1 != NULL) {
						r1->sprite.dx = (r1->sprite.dx + info->sprite.dx) >> 1;
						r1->sprite.dy = (r1->sprite.dy + info->sprite.dy) >> 1;
					}
					if (r2 != NULL) {
						r2->sprite.dx = (r2->sprite.dx + info->sprite.dx) >> 1;
						r2->sprite.dy = (r2->sprite.dy + info->sprite.dy) >> 1;
					}
					if (r3 != NULL) {
						r3->sprite.dx = (r3->sprite.dx + info->sprite.dx) >> 1;
						r3->sprite.dy = (r3->sprite.dy + info->sprite.dy) >> 1;
					}
					effect = effect_new(EFFECT_ID_EXPLOSION);
					sound_effect(SE_BOMB0);
				}
				break;

			default:
				effect = effect_new(EFFECT_ID_EXPLOSION);
				sound_effect(SE_BOMB0);
				break;
			}

			if (effect != NULL) {
				effect_set_pos(effect, &info->sprite, 0, 0);
			}
		}
	}

	return score;
}
