/*
 * shot.c
 *
 * Last modified: <2009/11/23 03:36:40 +0900> By SUMIDA
 */
#include <stdlib.h>
#include <math.h>
#include <SDL.h>
#include "system.h"
#include "texture.h"
#include "surface.h"
#include "sprite.h"
#include "effect.h"
#include "shot.h"

typedef struct SHOT_TYPE {
	int tex_id;
	int strength;
	Uint32 options;
	int frame_start;
	int frame_end;
	int frame_interval;
} SHOT_TYPE;

SHOT_TYPE shot_type[] = {
	{TEX_SHOT    ,  1,   0|SPRITE_FRAME, 0, 0, 0},
	{TEX_BALLET  ,  1,   0|SPRITE_FRAME8|SPRITE_USE_ANIME|SPRITE_LOOP_ANIME, 0, 3, 2},
	{TEX_MBALLET ,  2,   0|SPRITE_FRAME8|SPRITE_USE_ANIME|SPRITE_LOOP_ANIME, 0, 3, 2},
	{TEX_RAY1    ,  2,   0|SPRITE_FRAME, 0, 0, 0},
	{TEX_RAY2    ,  5, 100|SPRITE_FRAME, 0, 0, 0},
	{TEX_RAY3    ,  8, 100|SPRITE_FRAME, 0, 0, 0},
	{TEX_PHOTON  , 10,   0|SPRITE_FRAME, 0, 0, 0},
	{TEX_PHOTON_S, 50,   0|SPRITE_FRAME, 0, 0, 0},
	{TEX_PHOTON_S,  5,   0|SPRITE_FRAME, 0, 0, 0},
};

static void photon(SHOT_INFO* info);
static void photon_s(SHOT_INFO* info);

void(*func_shot[])(SHOT_INFO*) = {
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	photon,
	photon_s,
	photon_s,
};

void shot_init(SHOT_MGR *mgr, SHOT_INFO *infos, size_t number) {

	size_t i;

	if (mgr != NULL) {
		memset(mgr, 0, sizeof(SHOT_MGR));
		mgr->pool = infos;
	}

	memset(infos, 0, sizeof(SHOT_INFO) * number);

	for (i = 0; i < number - 1; i++) {
		infos[i].next = &infos[i + 1];
	}
}

SHOT_INFO *shot_new(SHOT_MGR *mgr, SHOT_ID id) {

	SHOT_INFO *info = mgr->pool;

	if (info != NULL) {
		SHOT_TYPE *p = &shot_type[id];
		SURFACE *surface = surface_get(p->tex_id);

		mgr->pool   = info->next;
		info->next  = mgr->active;
		mgr->active = info;

		info->id       = id;
		info->stat     = SHOT_STAT_NORMAL;
		info->count    = 0;
		info->strength = p->strength;
		info->sprite.surface_id = p->tex_id;
		info->sprite.options = SPRITE_VISIBLE|SPRITE_USE_MOVE|SPRITE_USE_COLLISION|p->options;
		info->sprite.frame_start    = p->frame_start;
		info->sprite.frame_end      = p->frame_end;
		info->sprite.frame_interval = p->frame_interval;
		info->sprite.collision.x0 = IntToReal(0);
		info->sprite.collision.y0 = IntToReal(0);
		info->sprite.collision.x1 = IntToReal(surface->frame_width - 1);
		info->sprite.collision.y1 = IntToReal(surface->frame_height - 1);
	}

	return info;
}

void shot_delete(SHOT_MGR *mgr, SHOT_INFO *info, SHOT_INFO *prev) {

	if (prev == NULL) {
		mgr->active = info->next;
	} else {
		prev->next = info->next;
	}

	memset(info, 0, sizeof(SHOT_INFO));
	info->next = mgr->pool;
	mgr->pool = info;
}

int shot_action(SHOT_MGR *mgr, SHIP_MGR *ship_mgr) {

	SHOT_INFO *prev = NULL;
	SHOT_INFO *info = mgr->active;
	SHIP_INFO *ship;

	int score = 0;

	while (info != NULL) {

		SPRITE *sprite = &info->sprite;

		/* Vbg̓ꏈ */
		if (func_shot[info->id] != NULL) {
			func_shot[info->id](info);
		}

		/* Vbg̈ړ */
		sprite_action(&info->sprite);

		/* ͂ݏoꍇ͍폜 */
		{
			SURFACE *surface = surface_get(info->sprite.surface_id);
			int x0 = RealToInt(info->sprite.x);
			int y0 = RealToInt(info->sprite.y);
			int x1 = x0 + surface->frame_width  -1;
			int y1 = y0 + surface->frame_height -1;

			if (x0 > CLIP_X1 || x1 < CLIP_X0 || y0 > CLIP_Y1 || y1 < CLIP_Y0) {
				SHOT_INFO *next = info->next;
				shot_delete(mgr, info, prev);
				info = next;
				continue;
			}
		}

		/* G@Ƃ̐ڐG */
		for (ship = ship_mgr->active; ship != NULL; ship = ship->next) {

			if (ship->stat != SHIP_STAT_LOST
				&& ship->stat != SHIP_STAT_EXPLOSION
				&& sprite_collision_check(&info->sprite, &ship->sprite) == SDL_TRUE) {

				/*REPORT("hit!\n");*/
				score += ship_down(ship, &info->strength, info->sprite.dx >> 3, info->sprite.dy >> 3);

				/* }烍bNIO */
				if (info->target != NULL) {
					info->target->flag &= ~SHIP_FLAG_LOCKON;
				}

				if (info->strength <= 0) {
					switch (info->id) {
					case SHOT_ID_NORMAL:
						{
							EFFECT_INFO *effect = effect_new(EFFECT_ID_HIT);

							if (effect != NULL) {
								effect_set_pos(effect, &info->sprite, 0, 0);
								effect->sprite.dx = ToReal(Random(-1.5, 1.5));
								effect->sprite.dy = ToReal(Random(0, 2));
							}
							break;
						}

					default:
						effect_set_explosion(&info->sprite);
						break;
					}
					{
						SHOT_INFO *next = info->next;
						shot_delete(mgr, info, prev);
						info = next;
						sprite = NULL;
						/*REPORT("shot delete!\n");*/
					}
					break;
				}
			}
		}

		/* ڐG̏ꍇ͕\ */
		if (sprite != NULL) {
			sprite_put(sprite);
			prev = info;
			info = info->next;
		}
	}

	return score;
}

void shot_fire(SHOT_INFO *info, SHIP_INFO *ship, Real ox, Real oy) {

	SURFACE *surface_shot = surface_get(info->sprite.surface_id);
	SURFACE *surface_ship = surface_get(ship->sprite.surface_id);

	info->sprite.x = ship->sprite.x + ox
		+ (IntToReal(surface_ship->frame_width - surface_shot->frame_width) >> 1);
	info->sprite.y = ship->sprite.y + oy
		+ (IntToReal(surface_ship->frame_height - surface_shot->frame_height) >> 1);
}

void shot_homing(SHOT_INFO *info, SHIP_INFO *ship, Real speed) {

	SURFACE *surface_ship = surface_get(ship->sprite.surface_id);
	SURFACE *surface_shot = surface_get(info->sprite.surface_id);
	double dx = (ship->sprite.x + (IntToReal(surface_ship->frame_width) >> 1)
				 - info->sprite.x - (IntToReal(surface_shot->frame_width) >> 1)) / 65536.0;
	double dy = (ship->sprite.y + (IntToReal(surface_ship->frame_height) >> 1)
				 - info->sprite.y - (IntToReal(surface_shot->frame_height) >> 1)) / 65536.0;
	double d = sqrt(dx * dx + dy * dy);

	if (d) {
		double spd = speed / 65536.0;
		info->sprite.dx = (Real)((dx / d * spd) * 65536);
		info->sprite.dy = (Real)((dy / d * spd) * 65536);
	} else {
		info->sprite.dx = 0;
		info->sprite.dy = speed;
	}
}

/* z[~Oq */
static void photon(SHOT_INFO* info) {

	if (info->target == NULL) return;

	if (info->count++ < 60 * 3) {

		Real dx = info->sprite.dx;
		Real dy = info->sprite.dy;

		shot_homing(info, info->target, IntToReal(3));

		info->sprite.dx = (info->sprite.dx + dx * 7) >> 3;
		info->sprite.dy = (info->sprite.dy + dy * 7) >> 3;

	} else if(!(info->count & 0x03)) {
		if (RealMul(info->sprite.dx, info->sprite.dx)
			+ RealMul(info->sprite.dy, info->sprite.dy) < IntToReal(32/*16*/)) {
			info->sprite.dx += info->sprite.dx >> 4;
			info->sprite.dy += info->sprite.dy >> 4;
		}
	}

	{
		EFFECT_INFO *effect = effect_new(EFFECT_ID_PHOTON_SHADOW);

		if (effect != NULL) {
			effect_set_pos(effect, &info->sprite, 0, 0);
		}
	}
}

/* ^z[~Oq */
static void photon_s(SHOT_INFO* info) {

	Real dx = info->sprite.dx;
	Real dy = info->sprite.dy;

	if (info->target != NULL) {
		shot_homing(info, info->target, IntToReal(8));

		info->sprite.dx = (info->sprite.dx + dx * 7) >> 3;
		info->sprite.dy = (info->sprite.dy + dy * 7) >> 3;

	} else {
		Real spd = RealMul(info->sprite.dx, info->sprite.dx)
			+ RealMul(info->sprite.dy, info->sprite.dy);

		if (spd < IntToReal(32)) {
			/* xSȂꍇ́AOɌĉ */
			info->sprite.dx = dx * 4 / 5;
			info->sprite.dy = (IntToReal(-8) + dy * 4) / 5;

		} else if (spd < IntToReal(50)) {
			/* xȂꍇ́A */
			info->sprite.dx = dx * 5 >> 2;
			info->sprite.dy = dy * 5 >> 2;
		}
	}

	{
		EFFECT_INFO *effect = effect_new(EFFECT_ID_PHOTON_S_SHADOW);

		if (effect != NULL) {
			effect_set_pos(effect, &info->sprite, 0, 0);
		}
	}
}

void shot_3way(SHIP_INFO *info, SHOT_MGR *shot_mgr, Real ox, Real oy, double angle, Real speed) {

	SHOT_INFO *shot1 = shot_new(shot_mgr, SHOT_ID_BALLET);
	SHOT_INFO *shot2 = shot_new(shot_mgr, SHOT_ID_BALLET);
	SHOT_INFO *shot3 = shot_new(shot_mgr, SHOT_ID_BALLET);

	if (shot1 != NULL) {
		shot_fire(shot1, info, ox, oy);
		sprite_set_angle(&shot1->sprite, angle, speed);
	}
	if (shot2 != NULL) {
		shot_fire(shot2, info, ox, oy);
		sprite_set_angle(&shot2->sprite, angle - 15, speed);
	}
	if (shot3 != NULL) {
		shot_fire(shot3, info, ox, oy);
		sprite_set_angle(&shot3->sprite, angle + 15, speed);
	}
}

void shot_5way(SHIP_INFO *info, SHOT_MGR *shot_mgr, Real ox, Real oy, double angle, Real speed) {

	SHOT_INFO *shot1 = shot_new(shot_mgr, SHOT_ID_BALLET);
	SHOT_INFO *shot2 = shot_new(shot_mgr, SHOT_ID_BALLET);
	SHOT_INFO *shot3 = shot_new(shot_mgr, SHOT_ID_BALLET);
	SHOT_INFO *shot4 = shot_new(shot_mgr, SHOT_ID_BALLET);
	SHOT_INFO *shot5 = shot_new(shot_mgr, SHOT_ID_BALLET);

	if (shot1 != NULL) {
		shot_fire(shot1, info, ox, oy);
		sprite_set_angle(&shot1->sprite, angle, speed);
	}
	if (shot2 != NULL) {
		shot_fire(shot2, info, ox, oy);
		sprite_set_angle(&shot2->sprite, angle - 15, speed);
	}
	if (shot3 != NULL) {
		shot_fire(shot3, info, ox, oy);
		sprite_set_angle(&shot3->sprite, angle + 15, speed);
	}
	if (shot4 != NULL) {
		shot_fire(shot4, info, ox, oy);
		sprite_set_angle(&shot4->sprite, angle - 30, speed);
	}
	if (shot5 != NULL) {
		shot_fire(shot5, info, ox, oy);
		sprite_set_angle(&shot5->sprite, angle + 30, speed);
	}
}

void shot_chack_target(SHOT_MGR *mgr, SHIP_INFO *ship) {

	SHOT_INFO *prev = NULL;
	SHOT_INFO *info = mgr->active;

	while (info != NULL) {

		if (ship == info->target) {
			info->target = NULL;
		}

		prev = info;
		info = info->next;
	}
}

/*
void shot_lost_target(SHOT_MGR *mgr, SHIP_INFO *ship) {

	SHOT_INFO *prev = NULL;
	SHOT_INFO *info = mgr->active;

	while (info != NULL) {

		if (ship == info->target) {

			SHOT_INFO *next = info->next;
			shot_delete(mgr, info, prev);
			info = next;

		} else {
			prev = info;
			info = info->next;
		}
	}
}
*/
