/*
 * system.h
 *
 * Last modified: <2009/11/28 22:15:22 +0900> By SUMIDA
 */
#ifndef SYSTEM_H_INCLUDED
#define SYSTEM_H_INCLUDED

#define WIDTH           640
#define HEIGHT          480
#define DEPTH            32
#define DELAY            40

#define CHANNEL_MAX      16
#define CHUNK_MAX        16

#define SHIP_MAX          3
#define ENEMY_MAX        50
#define SHIP_SHOT_MAX    10
#define ENEMY_SHOT_MAX  2000

#define EFFECT_MAX      200
#define LEVEL_MAX        15

#define RANKUP_RATE      4/5

#define CLIP_X0         -32
#define CLIP_Y0         -32
#define CLIP_X1         (WIDTH+32-1)
#define CLIP_Y1         (HEIGHT+32-1)

#define SYSTEM_OPT_VERBOSE    0x10000000
#define SYSTEM_OPT_FULLSCREEN 0x20000000
#define SYSTEM_OPT_LEVEL      0x00000010
#define SYSTEM_OPT_LEVEL_TYPE 0x0000000F
#define SYSTEM_OPT_EASY       0x00000000
#define SYSTEM_OPT_NORMAL     0x00000002
#define SYSTEM_OPT_HARD       0x00000006
#define SYSTEM_OPT_SUPER      0x0000000A
#define SYSTEM_OPT_HYPER      0x0000000E

#define SYSTEM_MODE_MASK      0x00000007
#define SYSTEM_MODE_READY     0x00000001
#define SYSTEM_MODE_RANKING   0x00000002
#define SYSTEM_MODE_SELECT    0x00000003
#define SYSTEM_MODE_OVER      0x00000004
#define SYSTEM_MODE_ENTRY     0x00000005
#define SYSTEM_MODE_SCORE     0x00000006
#define SYSTEM_MODE_PAUSE     0x00000007

#define SYSTEM_STAT_BOSS1     0x00010000
#define SYSTEM_STAT_BOSS2     0x00020000
#define SYSTEM_STAT_BOSS      (SYSTEM_STAT_BOSS1|SYSTEM_STAT_BOSS2)

/* VXe */
typedef struct SYSTEM_INFO {
	Uint32 options;
	Uint32 score;
	Uint32 status;
	Uint32 frame;
	int    level;
	int    count;
	int    encounter;
	int    map_index;
} SYSTEM_INFO;

extern SYSTEM_INFO system_info;

/* [U[ */
typedef enum LASER_TYPE {
	LASER_TYPE_SWORD = 0,
	LASER_TYPE_SEARCH,
	LASER_TYPE_LOCKON,
	LASER_TYPE_HOMING,
	LASER_TYPE_MAX
} LASER_TYPE;

#define NAME_LENGTH 10
/* vC[ */
typedef struct PLAYER_INFO {
	Uint32 score;
	LASER_TYPE laser_type;
	char start_level;
	char finish_level;
	char name[NAME_LENGTH + 1];
} PLAYER_INFO;

extern PLAYER_INFO player_info;

#define NUMBEROF(_array) (sizeof(_array)/sizeof(_array[0]))

#define C_DEG_RAD(_deg) ((_deg) * M_PI / 180.0)
#define C_RAD_DEG(_rad) ((_rad) * 180.0 / M_PI)

/* lower ȏ upper ȉ̗Ԃ */
extern double Random(double lower, double upper);

/* Pʉ~̍WpxZo */
extern double angle(double x, double y);

/* 擾 */
extern char *format(const char *format, ...);

/* |[go͂ */
extern void report(const char *file, const int line, const char *message);


#ifdef DEBUG
#define REPORT(_msg) report(__FILE__, __LINE__, _msg)
#else
#define REPORT(_msg) /**/
#endif

#endif
